import numpy as np
import random


class Alloy():
    """Alloy object, based on geometry"""
    def __init__(self,g):
        """Initialize the object"""
        self.dimensionality = g.dimensionality
        self.r = g.r.copy()
        self.n = len(self.r)
        self.nrep = 1
        self.a1 = g.a1
        self.a2 = g.a2
        self.a3 = g.a3
        self.specie = [0 for ri in self.r] # specie index
        self.nspecie = 1 # number of species
        self.setup_distances()
        self.fenergy = lambda d,i,j: 0.0
        self.setup_interaction()
    def setup_distances(self,**kwargs):
        self.d2 = setup_distances(self,**kwargs) # store square distances
    def setup_interaction(self,f=None):
        """Setup the objects required to compute interactions"""
        if type(f) is dict: f = dict2array(f,ns=self.nspecie)
        if f is None:
            self.get_energy_i = lambda ii: get_energy_i(self,self.fenergy,ii)
        elif callable(f): # input is a function
            self.get_energy_i = lambda ii: get_energy_i(self,f,ii)
        else:
            self.get_energy_i = lambda ii: wrapper_get_energy_i(self,ii,f)
    def get_energy(self):
        return get_energy(self)
    def set_species(self,s):
        self.specie = np.array(s) # specie index
        self.nspecie = len(np.unique(self.specie)) # number of species
    def minimize_energy(self,**kwargs):
        minimize_energy(self,**kwargs)
    def supercell(self,n):
        supercell(self,n)
    def write(self):
        write(self)
    def randomize(self,sp):
        """Randomize species"""
        ss = [0 for i in range(len(self.r))]
        n = len(self.r) # number of sites
        if sum(sp)!=n: raise
        if len(sp)!=2: raise
        ones = random.sample(range(0,n),sp[1])
        for o in ones: ss[o] = 1
        self.set_species(ss)

     


def setup_distances(self,n=1,cut=64):
    """Compute all the possible distances"""
    ds = np.array([0.0])
    r = self.r
    for i in range(-n,n+1):
      for j in range(-n,n+1):
        for k in range(-n,n+1):
            dn = []
            for ri in r:
              for rj in r:
                rk = rj + i*self.a1 + j*self.a2 + k*self.a3
                d = ri-rk
                d = d.dot(d)
#                d = np.array([d])
                dn.append(d)
            ds = np.concatenate([ds,dn]) # store
            ds = np.unique(ds) # retain unique
    ds = np.sort(ds)
    ds = ds[ds<cut]
    return ds # return distances



def get_energy(self,**kwargs):
    """Compute all the possible distances"""
    eout = 0.0
    r = self.r
    for ii in range(len(r)):
      eout += self.get_energy_i(ii,**kwargs)
    return eout 

def get_energy_i(self,f,ii,**kwargs):
    eout = 0.0
    r = self.r
    print(len(r)); exit()
    for jj in range(len(r)):
        eout += get_energy_ij(self,f,ii,jj,**kwargs)
    return eout



def get_energy_ij(self,f,ii,jj,n=1):
    etot = 0.0
    s = self.specie
    r = self.r
    ri = r[ii]
    rj = r[jj]
    si = s[ii]
    sj = s[jj]
    for i in range(-n,n+1):
      for j in range(-n,n+1):
        for k in range(-n,n+1):
                rk = rj + i*self.a1 + j*self.a2 + k*self.a3
                d = ri-rk
                d = d.dot(d)
                etot += f(d,si,sj) # function giving the energy
    return etot # return distances

def different_random_index(self):
    while True:
       ii = np.random.randint(0,self.n)
       jj = np.random.randint(0,self.n)
       if self.specie[ii]!=self.specie[jj]: return (ii,jj)


def minimize_energy(self,n=1000,mode="random",**kwargs):
    """Minimize the energy"""
    if mode=="random":
      eold = self.get_energy()
      for i in range(n):
          (ii,jj) = different_random_index(self)
          eold = single_update(self,eold,ii,jj,**kwargs)
    elif mode=="brute":
        def itene():
          eold = self.get_energy()
          for i in range(self.n):
            for j in range(self.n):
                if self.specie[i]==self.specie[j]: continue # next one
                eold = single_update(self,eold,i,j,**kwargs) # compute
          return eold
        eold = self.get_energy()
        while True:
            enew = itene() # iterate
            if abs(eold-enew)<1e-2: break
            eold = enew
    else: raise # not recognized

def single_update(self,eold,ii,jj,T=1e-7):
        s = self.specie.copy()
        sold = self.specie.copy()
        s[ii] = sold[jj]
        s[jj] = sold[ii]
        eijold = self.get_energy_i(ii) + self.get_energy_i(jj)
        self.specie = s # rewrite
        eijnew = self.get_energy_i(ii) + self.get_energy_i(jj)
        de = (eijnew - eijold)*2
        e = eold + de
        p = np.random.random()
        if de<0.0 or np.exp(-de/T)>p: # accept
            eold = e
            sold = s
            self.specie = s
            print("New energy ",e)
        else: self.specie = sold # next try
        return eold


from numba import jit

@jit(nopython=True)
def fn_get_energy_i(r,s,n,a1,a2,a3,ii,em,dis):
    """Specialized function for first neighbor interaction only"""
    etot = 0.0
    nn = len(dis) # number of neighbors
    dmax = np.max(dis) # maximum distance
    ri = r[ii]
    si = int(s[ii])
    for jj in range(len(r)): # loop over positions
      rj = r[jj]
      sj = int(s[jj])
      for i in range(-n,n+1):
        for j in range(-n,n+1):
          for k in range(-n,n+1):
                  rk = rj + i*a1 + j*a2 + k*a3
                  d = ri-rk
                  dd = np.sum(d*d)
                  if (dd-dmax)>1e-4: continue # too far
                  for nni in range(nn): # loop
                      if abs(dd-dis[nni])<1e-4: # this neighbor
                        etot = etot+ em[nni][si][sj] 
    return etot # return distances


def wrapper_get_energy_i(self,ii,f):
  f = np.array(f) # to array
  if len(f.shape)==2: f = np.array([f]) # redefine
  nn = len(f) # number of neighbors
  return fn_get_energy_i(self.r,self.specie,
                    self.nrep,self.a1,self.a2,self.a3,ii,f,
                    self.d2[1:nn+1])




def supercell(self,n):
    ro = []
    so = []
    for i in range(n):
      for j in range(n):
        for k in range(n):
            for ii in range(len(self.r)):
                ri = self.r[ii] +i*self.a1 + j*self.a2 + k*self.a3
                ro.append(ri)
                so.append(self.specie[ii])
    self.r = np.array(ro)
    self.a1 = n*self.a1
    self.a2 = n*self.a2
    self.a3 = n*self.a3
    self.n = len(self.r)
    self.specie = np.array(so)
    self.nspecie = len(self.specie)
    self.setup_distances()




def write(self,scale=1.0):
    r = self.r/scale
    np.savetxt("ALLOY.OUT",np.array([r[:,0],r[:,1],r[:,2],self.specie]).T)
    f = open("ATOMS.OUT","w")
    for i in range(len(r)):
        s = "S"+str(int(self.specie[i]))
        f.write(s+"   ")
        for ix in r[i]: f.write(str(ix)+"  ")
        f.write("\n")
    f.close()



def dict2array(d,ns=None):
    """Transform a dictionary into an array"""
    nn = max([di[2] for di in d]) # number of neighbors
    if ns is None:
      ns = max([di[0] for di in d]) # number of neighbors
    m = [np.zeros((ns,ns)) for i in range(nn)] # initialize
    for di in d:
        m[di[2]-1][di[0],di[1]] = d[di] # store value
        m[di[2]-1][di[1],di[0]] = d[di] # store value
    return np.array(m)


def array2dict(v):
    d = dict()
    for ni in range(len(v)):
        for i in range(len(v[ni])):
          for j in range(len(v[ni])):
              d[(i,j,ni+1)] = v[ni][i][j]
    return d






import numpy as np
from .. import geometry

def doped_MoS2():
    """Hamiltonian of doped MoS2"""
    g = geometry.honeycomb_lattice() # geometry
    h = g.get_hamiltonian(has_spin=True)
    h.add_sublattice_imbalance(0.6)
    h.add_soc(0.05)
    h.set_filling(0.501,nk=20)
    return h





import os
import numpy as np
from . import sculpt
from . import hamiltonians
from . import geometry



def generate_mullen_island(nx,ny,pristine=False):
  """Generate a Mullen island"""
  g = geometry.honeycomb_zigzag_ribbon(ny)
  g = g.supercell(2*nx+1+2)
  g.dimensionality = 0
  minx = np.min(g.x)
  
  def finite(x,y):
    if x<minx+.1: return False
    return True
  
  g = sculpt.intersec(g,finite)
  
  g = sculpt.remove_unibonded(g)
  
  
  g.center()
  
  # now add the hexagon
  
  dy = max(g.y) + 1.9
  
  r1 = np.array([-1.,0.+dy,0.])
  r2 = np.array([1.,0.+dy,0.])
  r3 = np.array([.5,np.sqrt(3.0)/2.+dy,0.])
  r4 = np.array([-.5,np.sqrt(3.0)/2.+dy,0.])
  r5 = np.array([.5,-np.sqrt(3.0)/2.+dy,0.])
  r6 = np.array([-.5,-np.sqrt(3.0)/2.+dy,0.])

  if not pristine:  
    g.r = [ri for ri in g.r] + [r1,r2,r3,r4,r5,r6]
    g.r = [ri for ri in g.r] + [-r1,-r2,-r3,-r4,-r5,-r6]
  #g.r =  [r1,r2,r3,r4,r5,r6]
  g.r = np.array(g.r)
  g.r2xyz()
  
  g.write()
  
  h = g.get_hamiltonian() # initialize hamiltonian
  
  def funh(ri,rj):
    dr = ri-rj
    if .1<dr.dot(dr)<1.3: return 1.0
    else: return 0.0
  
  
  h = hamiltonians.generate_parametric_hopping(h,funh)
  return h


def generate_mullen_ribbon(nx,ny,single=False):
  """Generate a Mullen island"""
  g = geometry.honeycomb_zigzag_ribbon(ny)
  g = g.supercell(2*nx+2)
  
  
  # now add the hexagon
  
  dy = max(g.y) + 1.9
  dx = np.sqrt(3.)/4 +0.
  xx = np.array([dx,0.,0.]) 
  g.r = [ri+xx for ri in g.r] # displace whole
  r1 = np.array([-1.,0.+dy,0.])
  r2 = np.array([1.,0.+dy,0.])
  r3 = np.array([.5,np.sqrt(3.0)/2.+dy,0.])
  r4 = np.array([-.5,np.sqrt(3.0)/2.+dy,0.])
  r5 = np.array([.5,-np.sqrt(3.0)/2.+dy,0.])
  r6 = np.array([-.5,-np.sqrt(3.0)/2.+dy,0.])
  
  g.r = [ri for ri in g.r] + [r1,r2,r3,r4,r5,r6] # upper
  if not single: g.r = [ri for ri in g.r] + [-r1,-r2,-r3,-r4,-r5,-r6] # lower
  #g.r =  [r1,r2,r3,r4,r5,r6]
  g.r = np.array(g.r)
  g.r2xyz()
  
  g.write()
  
  h = g.get_hamiltonian() # initialize hamiltonian
  
  def funh(ri,rj):
    dr = ri-rj
    if .1<dr.dot(dr)<1.3: return 1.0
    else: return 0.0
  
  
  h = hamiltonians.generate_parametric_hopping(h,funh)
  return h


def generate_mullen_surface(nx,ny):
  """Generate a Mullen surface, return bulk hamiltonian and surface cell"""
  g = geometry.honeycomb_lattice_zigzag_cell()
  g = g.supercell([2*nx+2+1,1]) # two dimensional
  gb = g.copy() # copy geometry
  g.dimensionality = 1 # reduce dimensionality
  g = sculpt.remove_unibonded(g)
  gb = sculpt.remove_unibonded(gb)
  g.center()
  gb.center()
  
  # now add the hexagon
  
  dy = max(g.y) + 1.9
  dx = 0.
  xx = np.array([dx,0.,0.]) 
  g.r = [ri+xx for ri in g.r] # displace whole
  r1 = np.array([-1.,0.+dy,0.])
  r2 = np.array([1.,0.+dy,0.])
  r3 = np.array([.5,np.sqrt(3.0)/2.+dy,0.])
  r4 = np.array([-.5,np.sqrt(3.0)/2.+dy,0.])
  r5 = np.array([.5,-np.sqrt(3.0)/2.+dy,0.])
  r6 = np.array([-.5,-np.sqrt(3.0)/2.+dy,0.])
  g.r = [ri for ri in g.r] + [r1,r2,r3,r4,r5,r6] # upper
  g.r = np.array(g.r)
  g.r2xyz()
  
  g.write()
  
  h = g.get_hamiltonian() # initialize hamiltonian

  def funh(ri,rj):
    dr = ri-rj
    if .1<dr.dot(dr)<1.3: return 1.0
    else: return 0.0
  
  
  h = hamiltonians.generate_parametric_hopping(h,funh)
  h.remove_spin() # remove spin degree of freedom 
  
  hb = gb.get_hamiltonian() # bulk hamiltonian
  hb.remove_spin() # remove spin
 
  return (h,hb) 





def generate_pentagon_island(nx,ny,pentagons=[0]):
  """Generate a Mullen island"""
  g = geometry.honeycomb_zigzag_ribbon(ny)
  g = g.supercell(2*nx+1)
  g.dimensionality = 0
  minx = np.min(g.x)
  
  def finite(x,y):
    if x<minx+.1: return False
    return True
  
  g = sculpt.intersec(g,finite)
  
  g = sculpt.remove_unibonded(g)
  
  
  g.center()
  
  # now add the hexagon
  
  dy = max(g.y) + 1.9

  ym = max(g.y) # position of zigzag atoms
  xz = [] # list of zigzag atoms

  for ir in g.r: # loop over atoms
    if (ir[1]-ym)<0.1: xz.append(ir[0]) # store upper zigzag atom
  for ip in pentagons: # loop over pentagons 
    # generate positions
    r1 = np.array([-1.+dx,0.+dy,0.])
    r2 = np.array([1.+dx,0.+dy,0.])
    r3 = np.array([.5+dx,np.sqrt(3.0)/2.+dy,0.])
    r4 = np.array([-.5+dx,np.sqrt(3.0)/2.+dy,0.])
    r5 = np.array([.5+dx,-np.sqrt(3.0)/2.+dy,0.])
    r6 = np.array([-.5+dx,-np.sqrt(3.0)/2.+dy,0.])
  
    # add the extra hexagons
    g.r = [ri for ri in g.r] + [r1,r2,r3,r4,r5,r6]
    g.r = [ri for ri in g.r] + [-r1,-r2,-r3,-r4,-r5,-r6]
  #g.r =  [r1,r2,r3,r4,r5,r6]
  g.r = np.array(g.r)
  g.r2xyz()
  
  g.write()
  
  h = g.get_hamiltonian() # initialize hamiltonian
  
  def funh(ri,rj):
    dr = ri-rj
    if .1<dr.dot(dr)<1.3: return 1.0
    else: return 0.0
  
  
  h = hamiltonians.generate_parametric_hopping(h,funh)
  return h












def real_singlet(h):
    """Given a Hamiltonian, return an operator that computes the
    real part of the spin singlet pairing"""
    op = h.copy()*0. # initialize
    op.add_swave(1.0) # add pairing
    from ..operators import Operator
    return Operator(op.intra) # return




from scipy.sparse import coo_matrix,bmat,csc_matrix
import numpy as np


# puts the matrix in spinor form
def m2spin(matin,matin2=[]):
  n=len(matin)
  from numpy import matrix
  if len(matin2)==0:
    matin2=matrix([[0.0j for i in range(n)] for j in range(n)])
  matout=matrix([[0.0j for i in range(2*n)] for j in range(2*n)])
  for i in range(n):
    for j in range(n):
      matout[2*i,2*j]=matin[i,j].copy()
      matout[2*i+1,2*j+1]=matin2[i,j].copy()
  return matout




def spinful(m,m2=None):
  """ Return a spinful hamiltonian"""
  if type(m)==type(np.matrix): return m2spin(m,matin2=m2)
  else: return spinful_sparse(m,m2=m2)


def spinful_sparse(m,m2=None):
  """ Return a spinful hamiltonian"""
  return m2spin_sparse(m,matin2=m2)


def m2spin_sparse(matin,matin2=None):
  m1 = coo_matrix(matin)
  if matin2 is None: m2 = m1
  else: m2 = coo_matrix(matin2)
  rows = np.concatenate([2*m1.row,2*m2.row+1])
  cols = np.concatenate([2*m1.col,2*m2.col+1])
  data = np.concatenate([m1.data,m2.data])
  dim = m1.shape[0]*2
  return csc_matrix((data,(rows,cols)),shape=(dim,dim), dtype=np.complex_)
  




def get_spinless2full(h,time_reversal=False,is_hamiltonian=True):
  """Function to transform a matrix into its full form"""
  if not h.has_spin and not h.has_eh: 
    def outf(m): return m # do nothing
  elif h.has_spin and not h.has_eh: # spinful and no eh
    def outf(m): 
        if time_reversal: return m2spin_sparse(m,np.conjugate(m)) # spinful
        else: return m2spin_sparse(m) # spinful
  elif h.check_mode("spinful_nambu"): # spinful and eh
    from .superconductivity import build_eh
    def outf(m): 
        if time_reversal: m2 = m2spin_sparse(m,np.conjugate(m)) # spinful
        else: m2 = m2spin_sparse(m) # spinful
        if is_hamiltonian: # use the Nambu transformation
            return build_eh(m2) # add e-h
        else: # just replicate the matrix (this is a temporal fix)
            return m2spin_sparse(m2) # replicate again
  elif h.check_mode("spinless_nambu"): 
      from .sctk import spinless
      return spinless.nambu
  else: raise
  return outf

  


def get_spinful2full(h):
  """Function to transform a matrix into its full form"""
  if not h.has_spin: raise
  else:
    if h.has_eh:
      from .superconductivity import build_eh
      def outf(m): 
        return build_eh(m) # add e-h
    else:
      def outf(m): return m # do nothing
  return outf







def des_spin(m,component=0):
    """Removes the spin degree of freedom"""
    from scipy.sparse import issparse
    sparse = issparse(m) # initial matrix is sparse
    m = coo_matrix(m) # convert
    n = m.shape[0]//2
    d1,r1,c1 = [],[],[]
    d0,r0,c0 = m.data,m.row,m.col
    for i in range(len(d0)):
        if r0[i]%2==component and c0[i]%2==component:
            d1.append(d0[i])
            r1.append(r0[i]//2)
            c1.append(c0[i]//2)
    mo = csc_matrix((d1,(r1,c1)),shape=(n,n),dtype=np.complex_)
    if not sparse: return mo.todense()
    else: return mo

from .htk.matrixcomponent import spin_mixing_part
from .htk.matrixcomponent import full2profile





def modify_hamiltonian_matrices(self,f0,use_geometry=False):
    """Apply a certain function to all the matrices"""
    # wrapper function
    if use_geometry: # use geometry
        g = self.geometry # get the geometry
        def f(m,dr): 
            return f0(m,g.r,g.replicas(d=dr)) # use r1 and r2
    else:
        def f(m,dr): return f0(m) # do not use geometry
    # apply the function
    self.intra = f(self.intra,[0,0,0]) # modify intracell
    if self.dimensionality==0: return  # zero dimensional systems
    self.turn_multicell() # multicell for all
    if self.is_multicell: # for multicell hamiltonians
      for i in range(len(self.hopping)): # loop over hoppings
        # modify Hamiltonian matrix
        self.hopping[i].m = f(self.hopping[i].m,self.hopping[i].dir) 
    else: # conventional way, now disabled
      raise



import numpy as np
from scipy.sparse import bmat,csc_matrix
import scipy.linalg as lg
from ..algebratk.unitary import make_unitary


def check_and_fix(smatrix,error=1e-7):
    """Given an smatrix as a list, chwck if it is Hermitian,
    and if not fix it"""
#    return smatrix
    n = smatrix[0][0].shape[0] # dimension of the matrix
    smatrix2 = [[csc_matrix(smatrix[i][j]) for j in range(2)] for i in range(2)]
    smatrix2 = bmat(smatrix2).todense()
    sH = np.conjugate(smatrix2).T
    merror = np.max(np.abs(lg.inv(smatrix2)-sH)) #  check unitarity
    if merror> error:
#        print("S-matrix is not unitary",error,"Determinant",np.abs(lg.det(sH)))
#        if abs(np.abs(lg.det(sH))-1.0)>1e-2: raise
#    print("S-matrix is unitary",error,"Determinant",np.abs(lg.det(sH)))
#    else: s3 = smatrix2
        smatrix2 = make_unitary(smatrix2)
#        print("Unitarized determinant",np.abs(lg.det(smatrix2)))
    s3 = np.matrix(smatrix2) # unitarized
    sout = [[s3[0:n,0:n],s3[n:2*n,0:n]],[s3[0:n,n:2*n],s3[n:2*n,n:2*n]]]
    return sout





import numpy as np
from . import supercell
from . import sculpt


def bulk2ribbon(obj,**kwargs):
    from .geometry import Geometry
    from .hamiltonians import Hamiltonian
    if type(obj)==Geometry:
        return geometry_bulk2ribbon(obj,**kwargs)
    elif type(obj)==Hamiltonian:
        return hamiltonian_ribbon(obj,**kwargs)
    else: raise



def geometry_bulk2ribbon(g,n=10,boundary=[1,0],clean=True):
  """Return the geometry of a ribbon"""
  go = g.copy() # copy
  m = [[boundary[0],boundary[1],0],[0,1,0],[0,0,1]] # supercell
  if boundary[0]!=1 or boundary[1]!=0:
    go = supercell.non_orthogonal_supercell(go,m,mode="fill") 
  go = go.supercell((1,n)) # create a supercell
  go = sculpt.rotate_a2b(go,go.a1,np.array([1.,0.,0.]))
  go.dimensionality = 1 # zero dimensional
  go.a2 = np.array([0.,1.,0.])
  if clean: 
    go = go.clean(iterative=True)
    if len(go.r)==0:
      print("Ribbon is not wide enough")
      raise
  go.real2fractional()
  go.fractional2real()
  go.celldis = go.a1[0]
  go.center()
  return go







def hamiltonian_ribbon(hin,n=10):
  """Return the Hamiltonian of a film"""
  h = hin.copy() # copy Hamiltonian
  from . import multicell
  h = multicell.supercell_hamiltonian(h,nsuper=[1,n,1])
  hopout = [] # list
  for i in range(len(h.hopping)): # loop over hoppings
    if abs(h.hopping[i].dir[2])<0.01: 
      if abs(h.hopping[i].dir[1])<0.01: 
        hopout.append(h.hopping[i])
  h.hopping = hopout
  if len(hopout)==0: raise # no hopping found
  h.dimensionality = 1
  h.geometry.dimensionality = 1
  h.geometry = sculpt.rotate_a2b(h.geometry,h.geometry.a1,np.array([1.,0.,0.]))
  return h





import numpy as np
import types


def hubbard_mf(EB,**kwargs):
    """Wrapper to perform a mean-field Hubbard calculation with an Embedding
    object."""
    ## This is just a workaround
    h = EB.H.copy() # copy Hamiltonian
    h.EB = EB.copy() # copy the object
    def dm(self,**kwargs):
        return {(0,0,0):self.EB.get_density_matrix(**kwargs)}
    def update(self,*args):
        self.EB.set_multihopping(*args)
    def fermi(self,*args,**kwargs): return 0.0
    def shift_fermi(self,*args,**kwargs): return None
    h.get_density_matrix = types.MethodType(dm,h) # overwrite
    h.set_multihopping = types.MethodType(update,h) # overwrite
    h.get_fermi4filling = types.MethodType(fermi,h) # overwrite
    h.shift_fermi = types.MethodType(shift_fermi,h) # overwrite
    h = h.get_mean_field_hamiltonian(**kwargs) # get the mean-field Hamiltonian
    return h.EB.copy() # return the new embedding object






from ..geometry import same_site
import numpy as np
from ..kanemele import get_haldane_function
from ..utilities import get_callable
from .dvector import dvector2delta


def pairing_generator(self,delta=0.0,mode="swave",d=[0.,0.,1.],
        nn = 1,
    **kwargs):
    """Create a generator, taking as input two positions, and returning
    the 2x2 pairing matrix"""
    # wrapper for the amplitude and d-vector
    deltaf = get_callable(delta) # callable for the amplitude
    df = get_callable(d) # callable for the d-vector
    if callable(mode):
        weightf = mode # mode is a function returning a 2x2 pairing matrix
    elif mode=="swave":
        weightf = lambda r1,r2: swave(r1,r2,H=self,nn=nn,**kwargs) 
    elif mode=="deltaud":
        # this is a workaround for delta_ud alone!
        weightf = lambda r1,r2: deltaud(r1,r2,deltaf) 
    elif mode=="extended_swave":
        weightf = lambda r1,r2: swave(r1,r2,H=self,nn=1,
                     **kwargs) #same_site(r1,r2)*np.identity(2)
    elif mode=="triplet": 
        weightf = lambda r1,r2: pwave(r1,r2,df,**kwargs)
    elif mode=="pwave": 
        weightf = lambda r1,r2: pwave(r1,r2,df,**kwargs)
    elif mode=="nodal_fwave":
        #weightf = lambda r1,r2: nodal_fwave(r1,r2,df,H=self,**kwargs)
        weightf = nodal_fwave_generator(df,H=self,**kwargs)
#    elif mode=="chiral_fwave": 
#        weightf = lambda r1,r2: get_triplet(r1,r2,df,L=3)
    elif mode=="chiral_pwave": 
        weightf = lambda r1,r2: get_triplet(r1,r2,df,L=1,**kwargs)
    elif mode=="chiral_fwave": 
        weightf = get_triplet_generator(df,L=3,H=self,nn=nn,**kwargs)
    elif mode=="chiral_dwave": 
        weightf = lambda r1,r2: get_singlet(r1,r2,L=2,**kwargs)
    elif mode=="chiral_gwave": 
        weightf = lambda r1,r2: get_singlet(r1,r2,L=4,**kwargs)
    elif mode=="antihaldane":
        f = get_haldane_function(self.geometry,stagger=True)
        weightf = lambda r1,r2: f(r1,r2)*np.identity(2)
    elif mode=="haldane":
        f = get_haldane_function(self.geometry,stagger=False)
        weightf = lambda r1,r2: f(r1,r2)*np.identity(2)
    elif mode=="swavez":
        weightf = lambda r1,r2: same_site(r1,r2)*tauz
    elif mode=="px":
        weightf = lambda r1,r2: px(r1,r2)
    elif mode=="dpid":
        weightf = lambda r1,r2: dpid(r1,r2,**kwargs)
    elif mode=="swaveA":
        weightf = lambda r1,r2: swaveA(self.geometry,r1,r2)
    elif mode=="swaveB":
        weightf = lambda r1,r2: swaveB(self.geometry,r1,r2)
    elif mode=="swavesublattice":
        def weightf(r1,r2):
          return swaveB(self.geometry,r1,r2) - swaveA(self.geometry,r1,r2)
    elif mode in ["dx2y2","nodal_dwave"]:
        weightf = lambda r1,r2: dx2y2(r1,r2,H=self,**kwargs)
    elif mode=="dxy":
        weightf = lambda r1,r2: dxy(r1,r2,H=self,**kwargs)
    elif mode=="snn":
        weightf = lambda r1,r2: swavenn(r1,r2)
    elif mode=="C3nn":
        weightf = lambda r1,r2: C3nn(r1,r2)
    elif mode=="SnnAB":
        weightf = lambda r1,r2: SnnAB(self.geometry,r1,r2)
    else: raise
    matrixf = lambda r1,r2: deltaf((r1+r2)/2.)*weightf(r1,r2) 
    return matrixf # return function

# matrices for the e-h subsector
iden = np.array([[1.,0.],[0.,1.]],dtype=np.complex_)
tauz = np.array([[1.,0.],[0.,-1.]],dtype=np.complex_)
taux = np.array([[0.,1.],[1.,0.]],dtype=np.complex_)
tauy = np.array([[0.,1j],[-1j,0.]],dtype=np.complex_)
UU = taux + 1j*tauy # projector in the UU sector
DD = taux - 1j*tauy # projector in DD sector


def swavenn(r1,r2):
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        return 1.0*iden
    return 0.0*iden


def dx2y2(r1,r2,**kwargs):
    """Function with first neighbor dx2y2 profile"""
    return (get_singlet(r1,r2,L=2,**kwargs) + get_singlet(r1,r2,L=-2,**kwargs))/2.
#    dr = r1-r2
#    dr2 = dr.dot(dr)
#    if 0.99<dr2<1.001: # first neighbor
#        return (dr[0]**2 - dr[1]**2)*iden
#    return 0.0*iden


def dxy(r1,r2,**kwargs):
    """Function with first neighbor dxy profile"""
    return (get_singlet(r1,r2,L=2,**kwargs) - get_singlet(r1,r2,L=-2,**kwargs))/2.
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        return (dr[0]*dr[1])*iden
    return 0.0*iden


def swave(r1,r2,nn=0,**kwargs):
    """Function with first neighbor dxy profile"""
    if nn==0: return same_site(r1,r2)*np.identity(2)
    else: return get_singlet(r1,r2,L=0,nn=nn,**kwargs)


def C3nn(r1,r2):
    """Function with first neighbor C3 profile"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
#        return dr[0]
        phi = np.arctan2(dr[1],dr[0]) # angle
        return 1.0*np.exp(1j*phi)*tauz
    return 0.0*tauz



def C3nn(r1,r2):
    """Function with first neighbor C3 profile"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
#        return dr[0]
        phi = np.arctan2(dr[1],dr[0]) # angle
        return 1.0*np.exp(1j*phi)*tauz
    return 0.0*tauz



def SnnAB(g,r1,r2):
    """Swave between AB"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: # first neighbor
        i = g.get_index(r1,replicas=True)
        j = g.get_index(r2,replicas=True)
        if g.sublattice[i]==1 and g.sublattice[j]==-1:
          return 1.0*tauz
        else: return -1.0*tauz
#          return 1.0*np.matrix([[1.,0.],[0.,0.]])
    return 0.0*iden




def swaveA(g,r1,r2):
    """Swave only in A"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if dr2<0.001: # first neighbor
        i = g.get_index(r1,replicas=False)
        if i is None: return 0.0
        if g.sublattice[i]==1:
          return 1.0*iden
    return 0.0*iden


def swaveB(g,r1,r2):
    """Swave only in A"""
    dr = r1-r2
    dr2 = dr.dot(dr)
    if dr2<0.001: # first neighbor
        i = g.get_index(r1,replicas=False)
        if i is None: return 0.0
        if g.sublattice[i]==-1:
          return 1.0*iden
    return 0.0*iden


def px(r1,r2):
    """Function with first neighbor px profile"""
    dr = r1-r2 ; dr2 = dr.dot(dr)
    if 0.99<dr2<1.001: return dr[0]*tauz
    return 0.0*tauz



def get_triplet_generator(df,nn=1,H=None,**kwargs):
    if nn>1: # more than first neighbor
      dist = H.geometry.get_neighbor_distances(n=nn)[nn-1] # get this distance
      dist2 = dist**2
    else: dist2 = 1.0 # first neighbor
    return lambda r1,r2: get_triplet(r1,r2,df,dist2=dist2,**kwargs)



def get_triplet(r1,r2,df,L=1,dist2=1.0):
    """Function for triplet order"""
    dr = r1-r2 ; dr2 = dr.dot(dr)
    if abs(L)%2!=1: raise
    if np.abs(dr2-dist2)<1e-4:
        phi = np.arctan2(dr[1],dr[0])
        d = df((r1+r2)/2.) # evaluate dvector
        delta = dvector2delta(d) # compute the local deltas
        ms = np.array([[delta[2],delta[0]],[delta[1],-delta[2]]])
        return np.exp(1j*phi*L)*np.array(ms,dtype=np.complex_)
    else: return 0.0*tauz


def get_singlet(r1,r2,L=2,phi0=0.,H=None,nn=1):
    """Function for p-wave order"""
    if L%2!=0: raise
    dr = r1-r2 ; dr2 = dr.dot(dr)
    if nn>1: # more than first neighbor
      dist = H.geometry.get_neighbor_distances(n=nn)[nn-1] # get this distance
      dist2 = dist**2
    else: dist2 = 1.0 # first neighbor
#      print(dist2) ; exit()
    if np.abs(dr2-dist2)<1e-4:
        phi = np.arctan2(dr[1],dr[0])
        out = np.exp(1j*(phi+phi0*np.pi*2.)*L)
#        print(out)
        return out*iden
    else: return 0.0*tauz


def get_deltaud(r1,r2,f):
    """Return a Delta ud, given a certain function f of the two positions"""
    return iden*f(r1,r2)



def pwave(*args,**kwargs): 
    return get_triplet(*args,L=1,**kwargs)


def dpid(*args,**kwargs):
    return get_singlet(*args,L=2,**kwargs)

def nodal_fwave_generator(*args,dphi=0.,**kwargs): 
    """Generator for real f-wave order"""
    z = np.exp(1j*np.pi*2*dphi) # complex rotation
    f1 = get_triplet_generator(*args,L=3,**kwargs)
    f2 = get_triplet_generator(*args,L=-3,**kwargs)
    return lambda r1,r2: f1(r1,r2) + z*f2(r1,r2)

def nodal_fwave(*args,dphi=0.,**kwargs): 
    z = np.exp(1j*np.pi*2*dphi) # complex rotation
    return get_triplet(*args,L=3,**kwargs) + z*get_triplet(*args,L=-3,**kwargs)



from .. import algebra
import numpy as np
from .. import green
from ..hamiltonians import is_number

dagger = algebra.dagger

# Landauer Buttiker formula

def landauer(HT,energy=0.0,error=1e-9,**kwargs):
    """ Calculates transmission using Landauer formula"""
    delta = HT.delta
    if not HT.block_diagonal:
      intra = HT.central_intra # central intraterm   
      dimhc = intra.shape[0] # dimension of the central part
    if HT.block_diagonal:
        if len(HT.central_intra)==0: # no central
            print("No central region provided")
            raise
        intra = HT.central_intra[0][0] # when it is diagonal
 # dimension of the central part
        dimhc = len(HT.central_intra)*intra.shape[0]
    iden = np.matrix(np.identity(len(intra),dtype=complex)) # create identity
    selfl = HT.get_selfenergy(energy,lead=0,delta=delta,pristine=False) # left Sigma
    selfr = HT.get_selfenergy(energy,lead=1,delta=delta,pristine=False) # right Sigma
    #################################
    # calculate Gammas 
    #################################
    gammar = 1j*(selfr-dagger(selfr))
    gammal = 1j*(selfl-dagger(selfl))
 
    #################################
    # dyson equation for the center
    #################################
    # central green function
    intra = HT.central_intra
    # full matrix
    if not HT.block_diagonal:
        heff = intra + selfl + selfr
        HT.heff = heff
        gc = (energy+1j*delta)*iden - heff
        gc = algebra.inv(gc) # calculate inverse
        G = np.trace(gammar@gc@gammal@dagger(gc)).real
        return G
    # reduced matrix
    if HT.block_diagonal:
        from copy import deepcopy
        heff = deepcopy(intra)
        heff[0][0] = intra[0][0] + selfl
        heff[-1][-1] = intra[-1][-1] + selfr
        dd = (energy+1j*delta)*iden
        for i in range(len(intra)):  # add the diagonal energy part
          heff[i][i] = heff[i][i] - dd  # this has the wrong sign!!
       # now change the sign
        for i in range(len(intra)):
          for j in range(len(intra)):
            if heff[i][j] is not None:
              heff[i][j] = -heff[i][j]
        # calculate green function
        gauss_inverse = green.gauss_inverse  # routine to invert the matrix
        # calculate only some elements of the central green function
        gcn1 = gauss_inverse(heff,len(heff)-1,0) # calculate element 1,n
        # and apply Landauer formula
        G = np.trace(gammar@gcn1@gammal@dagger(gcn1)).real
    return G # return transmission




import numpy as np


def merge_channels(h1,h2):
    """Merge the spin up and down channels"""
    if h1.has_spin or h2.has_spin: raise
    if h1.has_eh or h2.has_eh: raise
    h1 = h1.get_multicell() # multicell Hamiltonian
    h2 = h2.get_multicell() # multicell Hamiltonian
    if len(h1.hopping)!=len(h2.hopping): raise
    h = h1.copy() # copy Hamiltonian
    h.turn_spinful() # make spinful
    h.turn_dense()
    h1.turn_dense()
    h2.turn_dense()
    h.clean()
    if h.has_eh: raise # not implemented
    n = len(h.geometry.r) # number of sites
    for i in range(n): # loop over positions
      for j in range(n): # loop over positions
        h.intra[2*i,2*j] = h2.intra[i,j] # set that spin channel
        h.intra[2*i+1,2*j+1] = h1.intra[i,j] # set that spin channel
    for k in range(len(h.hopping)): # loop
        d = np.array(h.hopping[k].dir)
        d1 = np.array(h1.hopping[k].dir)
        d2 = np.array(h2.hopping[k].dir)
        dd1 = d-d1
        dd2 = d-d1
        if dd1.dot(dd1)>0.001 or dd2.dot(dd2)>0.001: raise
        for i in range(n): # loop over positions
          for j in range(n): # loop over positions
            h.hopping[k].m[2*i,2*j] = h2.hopping[k].m[i,j]
            h.hopping[k].m[2*i+1,2*j+1] = h1.hopping[k].m[i,j]
    return h







from ..heterostructures import create_leads_and_central_list
from ..heterostructures import Heterostructure
from ..algebra import dagger
import numpy as np


def build(left=None,right=None,central=None,**kwargs):
    """Create a heterostructure, works also for 2d"""
  #  if central is None: central = [h1,h2] # list
    if left is None or right is None:
        print("No leads provided")
        raise
    h1,h2 = left,right
    if central is None: central = [] # list
    # make the Hamiltonians compatible
    from ..htk.mode import make_compatible
    h1 = make_compatible(h1,h2)
    h2 = make_compatible(h2,h1)
    central = [make_compatible(hi,h1) for hi in central]   
    ###############
    if h1.dimensionality==h2.dimensionality: # same dimensionality
        if h1.dimensionality==1: # one dimensional
          return create_leads_and_central_list(h2,h1,central,**kwargs) # standard way
        elif h1.dimensionality==2:  # two dimensional
          def fun(k,lc=1.0,rc=1.0):
            # evaluate at a particular k point
            h1p = h1.get_1dh(k)
            h2p = h2.get_1dh(k)
            centralp = [hc.get_1dh(k) for hc in central]
            out = create_leads_and_central_list(h2p,h1p,centralp,**kwargs) # standard way
            out.scale_lc = lc
            out.scale_rc = rc
            return out # return 1d heterostructure
          hout = Heterostructure() # create
          hout.Hr = h2.copy() # store Hamiltonian
          hout.Hl = h1.copy() # store Hamiltonian
          hout.dimensionality = 2 # two dimensional
          hout.generate = fun # function that generates the heterostructure
          return hout # function that return a heterostructure
        else: raise # not implemented
    else:
  # this is a temporal fix
        if h1.dimensionality==1 and h2.dimensionality==2: # 1D to 2D
            if len(central)!=0: raise # not implemented
            h2t = h2.get_1dh(0.) # create a 1D Hamiltonian
            HT = build(h1,h2t,central=central,**kwargs) # create a fake HT
            HT0 = HT.copy() # copy the HT
            def get_selfenergy(energy,nk=40,lead=0,delta=HT.delta,**kwargs):
                if lead==0: return HT0.get_selfenergy(energy,lead=lead,
                                            delta=delta,
                                         **kwargs) # default
                elif lead==1:
                    from .. import green
                    gf = green.bloch_selfenergy(h2,energy=energy,
                                           nk=nk,mode="adaptive",
                                           delta=delta)[0]
                    gamma = dagger(HT0.left_inter)
                    return gamma@gf@dagger(gamma)
                else: raise
            HT.get_selfenergy = get_selfenergy # overwrite
            HT.block_diagonal = False # block diagonal
            HT.central_intra = h1.intra # intra cell
            HT.right_coupling = dagger(HT.left_coupling)
            return HT
        raise # not finished



class Hybrid_Heterostructure(Heterostructure):
    """New class for hybrid heterostructures"""
    def __init__(self,h1,h2):
        if h1.dimensionality==1 and h2.dimensionality==2: # 1D to 2D
          if len(central)!=0: raise # not implemented
          h2t = h2.get_1dh(0.) # create a 1D Hamiltonian
          HT = build(h1,h2t,central=central,**kwargs) # create a fake HT
          HT0 = HT.copy() # copy the HT
          def get_selfenergy(energy,nk=40,lead=0,delta=HT.delta,**kwargs):
              if lead==0: return HT0.get_selfenergy(energy,lead=lead,
                                          delta=delta,
                                       **kwargs) # default
              elif lead==1:
                  from .. import green
                  gf = green.bloch_selfenergy(h2,energy=energy,
                                         nk=nk,mode="adaptive",
                                         delta=delta)[0]
                  gamma = dagger(HT0.left_inter)
                  return gamma@gf@dagger(gamma)
              else: raise
          HT.get_selfenergy = get_selfenergy # overwrite
          HT.block_diagonal = False # block diagonal
          HT.central_intra = h1.intra # intra cell
          HT.right_coupling = dagger(HT.left_coupling)
          raise # not finished



def get_reflection_normal_lead(ht,s):
    """Get reflection matrix of the normal lead"""
    r1,r2 = s[0][0],s[1][1] # get the reflection matrices
    get_eh = ht.get_eh_sector # function to read either electron or hole
    # select the normal lead
    # r1 is normal
    if np.sum(np.abs(get_eh(ht.left_intra,i=0,j=1)))<0.0001: r = r1
    elif np.sum(np.abs(get_eh(ht.right_intra,i=0,j=1)))<0.0001: r = r2
    else:
        print("There is SC in both leads, aborting")
        raise
    return r



















import numpy as np
import scipy.linalg as lg
from . import algebra


from scipy.sparse import csc_matrix,bmat
from .spin import sx,sy,sz


def rotation_matrix(m,vectors):
  """ Rotates a matrix, to align its components with the direction
  of the magnetism """
  if not len(m)==2*len(vectors): # stop if they don't have
                                  # compatible dimensions
     raise
  # pauli matrices
  n = len(m)/2 # number of sites
  R = [[None for i in range(n)] for j in range(n)] # rotation matrix
  from scipy.linalg import expm  # exponenciate matrix
  for (i,v) in zip(range(n),vectors): # loop over sites
    vv = np.sqrt(v.dot(v)) # norm of v
    if vv>0.000001: # if nonzero scale
      u = v/vv
    else: # if zero put to zero
      u = np.array([0.,0.,0.,])
#    rot = u[0]*sx + u[1]*sy + u[2]*sz 
    uxy = np.sqrt(u[0]**2 + u[1]**2) # component in xy plane
    phi = np.arctan2(u[1],u[0])
    theta = np.arctan2(uxy,u[2])
    r1 =  phi*sz/2.0 # rotate along z
    r2 =  theta*sy/2.0 # rotate along y
    # a factor 2 is taken out due to 1/2 of S
    rot = expm(1j*r2) * expm(1j*r1)   
    R[i][i] = rot  # save term
  R = bmat(R)  # convert to full sparse matrix
  return R.todense()



def align_magnetism(m,vectors):
  """ Align matrix with the magnetic moments"""
  R = rotation_matrix(m,vectors) # return the rotation matrix
  mout = R * csc_matrix(m) * R.H  # rotate matrix
  return mout.todense() # return dense matrix





def global_spin_rotation(m,vector = np.array([0.,0.,1.]),angle = 0.0,
                             spiral = False,atoms = None):
  """ Rotates a matrix along a certain qvector """
  # pauli matrices
  from scipy.sparse import csc_matrix,bmat
  iden = csc_matrix([[1.,0.],[0.,1.]])
  n = m.shape[0]//2 # number of sites
  if atoms==None: 
    atoms = range(n) # all the atoms
  else: 
    raise
  R = [[None for i in range(n)] for j in range(n)] # rotation matrix
  for i in range(n): # loop over sites
      u = np.array(vector) # rotation direction
      u = u/np.sqrt(u.dot(u)) # normalize rotation direction
      rot = (u[0]*sx + u[1]*sy + u[2]*sz)/2. # rotation
      # a factor 2 is taken out due to 1/2 of S
      # a factor 2 is added to have BZ in the interval 0,1
      rot = algebra.todense(rot)
      rot = lg.expm(2.*np.pi*1j*rot*angle/2.0)
  #    if i in atoms:
      R[i][i] = rot  # save term
#    else:
#      R[i][i] = iden  # save term
  R = bmat(R)  # convert to full sparse matrix
  if spiral:  # for spin spiral
    mout = R @ m  # rotate matrix
  else:  # normal global rotation
    mout = R @ m @ algebra.dagger(R)  # rotate matrix
  return mout # return dense matrix




def spiralhopping(m,ri,rj,svector = np.array([0.,0.,1.]),
        qvector=[1.,0.,0.]): 
  """ Rotates a hopping matrix to create a spin spiral
  antsaz
      - ri and rj must be coordinates in lattice constants
      - svector is the axis of the rotation
      - qvector is the vector of the spin spiral
  """
  from scipy.sparse import csc_matrix,bmat
  iden = csc_matrix([[1.,0.],[0.,1.]]) # identity matrix
  def getR(r):
      """Return a rotation matrix"""
      n = len(r) # number of sites 
      R = [[None for i in range(n)] for j in range(n)] # rotation matrix
      u = np.array(svector) # rotation direction
      u = u/np.sqrt(u.dot(u)) # normalize rotation direction
      for i in range(n): # loop over sites
         rot = u[0]*sx + u[1]*sy + u[2]*sz 
         angle = np.array(qvector).dot(np.array(r[i])) # angle of rotation
         # a factor 2 is taken out due to 1/2 of S
         # a factor 2 is added to have BZ in the interval 0,1
         R[i][i] = algebra.expm(2.*np.pi*1j*rot*angle/2.0)
      return algebra.bmat(R)  # convert to full sparse matrix
  Roti = getR(ri) # get the first rotation matrix
  Rotj = getR(rj) # get the second rotation matrix
#  print(Roti@Rotj.H)
#  print(ri,rj)
  return Rotj @ m @ algebra.dagger(Roti) # return the rotated matrix


def hamiltonian_spin_rotation(self,vector=np.array([0.,0.,1.]),angle=0.):
    """ Perform a global spin rotation """
    if not self.has_spin: raise # no spin in the Hamiltonian
    gsr = global_spin_rotation # rename method
    if self.has_eh: raise
    self.intra = gsr(self.intra,vector=vector,angle=angle)
    if self.is_multicell: # multicell hamiltonian
      for i in range(len(self.hopping)): # loop 
        self.hopping[i].m = gsr(self.hopping[i].m,vector=vector,angle=angle)
    else:
      if self.dimensionality==0: pass
      elif self.dimensionality==1:
        self.inter = gsr(self.inter,vector=vector,angle=angle)
      elif self.dimensionality==2:
        self.tx = gsr(self.tx,vector=vector,angle=angle)
        self.ty = gsr(self.ty,vector=vector,angle=angle)
        self.txy = gsr(self.txy,vector=vector,angle=angle)
        self.txmy = gsr(self.txmy,vector=vector,angle=angle)
      else: raise



def generate_spin_spiral(self,vector=np.array([0.,0.,1.]),
                            qspiral=[1.,0.,0.],fractional=True,
                            **kwargs):
    """
    Generate a spin spiral antsaz in the Hamiltonian
    """
    if not self.has_spin: raise # no spin
    qspiral = np.array(qspiral) # to array
    if qspiral.dot(qspiral)<1e-7: qspiral = np.array([0.,0.,0.])
    self.geometry.get_fractional()
    def tmprot(m,vec): # function used to rotate
      """Function to rotate one matrix"""
      if fractional: # fractional coordinates provided
        # rotate fractional coordinates
        ri = self.geometry.frac_r # positions of the first cell
        rj = self.geometry.frac_r + np.array(vec) # positions of the next cell
        return spiralhopping(m,ri,rj,svector=vector,
                qvector = qspiral)
      else:
        # only rotate between supercells
        angleq = qspiral.dot(np.array(vec)) # angle of the rotation
        return global_spin_rotation(m,vector=vector,
              angle=angleq,spiral=True,atoms=None)
    self.intra = tmprot(self.intra,[0.,0.,0.]) # rotate intra matrix
    # now rotate every matrix
    if self.is_multicell: # multicell Hamiltonian
      a1,a2,a3 = self.geometry.a1, self.geometry.a2,self.geometry.a3
      for i in range(len(self.hopping)): # loop
        ar = self.hopping[i].dir # direction
#        direc = a1*ar[0] + a2*ar[1] + a3*ar[2]
        self.hopping[i].m = tmprot(self.hopping[i].m,ar) # rotate matrix
    else:
      if self.dimensionality==0: pass
      elif self.dimensionality==1:
        self.inter = tmprot(self.inter,[1.,0.,0.])
      elif self.dimensionality==2:
        a1,a2 = self.geometry.a1,self.geometry.a2
        self.tx = tmprot(self.tx,[1.,0.,0.])
        self.ty = tmprot(self.ty,[0.,1.,0.])
        self.txy = tmprot(self.txy,[1.,1.,0.])
        self.txmy = tmprot(self.txmy,[1.,-1.,0.])
      else: raise



# compute the kappa parameter of a heterostructure
import numpy as np
from ..parallel import pcall

def get_single(HT=None,c=1.0,energies=[0.0],**kwargs):
    """Get a single conductance"""
    HT.set_coupling(c) # scaling
    return np.array([HT.didv(energy=e,**kwargs) for e in energies]) # loop over Ts


def get_conductances(T=1e-2,**kwargs):
    """Compute Kappa by doing a log-log plot"""
    cref = T
    ts = np.exp(np.linspace(np.log(cref*0.9),np.log(cref*1.1),2)) # hoppings
#    ts = [cref*0.9,cref*1.1]
    Gs = np.array([get_single(c=t,**kwargs) for t in ts]) # compute conductance
    return ts,Gs

def get_power(ts,gs,delta=1e-8):
    """Given hoppings and conductances, extract the power"""
#    ts2 = ts[gs>delta]
#    gs2 = gs[gs>delta]
    p = np.polyfit(np.log(ts),np.log(gs),1)
    k = p[0]
    return k

def get_kappa(energy=0.0,**kwargs):
    ts,Gs = get_conductances(energies=[energy],**kwargs)
    ks = []
    for g in Gs.T: # loop over energies
        k = get_power(ts,g)
        ks.append(k)
    return np.array(ks)[0] # return kappa


def get_kappa_ratio(HT,**kwargs):
    ks1 = get_kappa(HT=generate_HT(HT,SC=True,**kwargs),**kwargs)
    ks2 = get_kappa(HT=generate_HT(HT,SC=False,**kwargs),**kwargs)
    return ks1/ks2


def generate_HT(ht,SC=True,**kwargs):
    """Given a heterostructure, generate a new one to compute kappa"""
    # this is a workaround
    from ..heterostructures import Heterostructure
    from .localprobe import LocalProbe
    from ..embeddingtk.kappa import DualLocalProbe
    def f(h):
        h = h.copy()
        if not SC: # remove the SC order
            h.remove_nambu()
            h.setup_nambu_spinor()
        return h
    if type(ht)==Heterostructure: # heterostructure type
        from ..heterostructures import build
        Hr = f(ht.Hr)
        Hl = f(ht.Hl)
        hto = build(Hl,Hr) # create a new heterostructure
        hto.delta = ht.delta
        return hto
    elif type(ht)==LocalProbe: # Localprobe type
        out = ht.copy() # make a copy
        out.H = f(out.H)
        out.lead = f(out.lead)
        return out
    elif type(ht)==DualLocalProbe: # Dual Localprobe object
        if SC: return ht.SC_LP
        else: return ht.normal_LP
    else: raise


#### These are workrounds for more efficient finite temperature calculations ##
## Not yet finished! ##

def get_kappa_finite_temperature_energies(**kwargs):
    """Compute kappa using temperature convolution"""
    ts,Gs = get_conductances_finite_temp(**kwargs) # G at T=0
    ks = []
    for g in Gs.T: # loop over energies
        k = get_power(ts,g)
        ks.append(k)
    return np.array(ks) # return kappa


def get_kappa_finite_temperature_energies(HT,**kwargs):
    ks1 = get_kappa_finite_temperature_energies(
                     HT=generate_HT(HT,SC=True,**kwargs),**kwargs)
    ks2 = get_kappa_finite_temperature_energies(
                     HT=generate_HT(HT,SC=False,**kwargs),**kwargs)
    return ks1/ks2




import numpy as np
from scipy.sparse import csc_matrix,identity
import scipy.sparse as sp

dense = False

def block2nambu_matrix(m):
    if dense: return block2nambu_matrix_dense(m)
    else: return block2nambu_matrix_sparse(m)



def block2nambu_matrix_sparse(m):
  '''Reorder a matrix that has electrons and holes, so that
  the order resembles the Nambu spinor in each site.
  The initial matrix is
  H D
  D H
  The output is a set of block matrices for each site in the 
  Nambu form'''
  nr = m.shape[0]//4 # number of positions
  col,row,data = [],[],[]
  for i in range(nr): # electrons
    col = col + [2*i,2*i+1,2*i+2*nr,2*i+1+2*nr]
    row = row + [4*i,4*i+1,4*i+2,4*i+3]
    data = data + [1.,1.,1.,1.]
  R = sp.coo_matrix((data,(row,col)),shape=m.shape,dtype=complex)
  return R.T



def block2nambu_matrix_dense(m):
  '''Reorder a matrix that has electrons and holes, so that
  the order resembles the Nambu spinor in each site.
  The initial matrix is
  H D
  D H
  The output is a set of block matrices for each site in the 
  Nambu form'''
  R = np.matrix(np.zeros(m.shape)) # zero matrix
  nr = m.shape[0]//4 # number of positions
  for i in range(nr): # electrons
    R[2*i,4*i] = 1.0 # up electron
    R[2*i+1,4*i+1] = 1.0 # down electron
    R[2*i+2*nr,4*i+2] = 1.0 # down holes
    R[2*i+1+2*nr,4*i+3] = 1.0 # up holes
  R = csc_matrix(R) # to sparse
  return R


def block2nambu(m):
    R = block2nambu_matrix(m)
    Rh = np.conjugate(R.T)
    return Rh@m@R

reorder = block2nambu

def nambu2block(m):
    R = block2nambu_matrix(m)
    Rh = np.conjugate(R.T)
    return R@m@Rh





import numpy as np

def jackson_kernel(mus):
  """ Modify coeficient using the Jackson Kernel"""
  mo = mus.copy() # copy array
  n = len(mo)
  pn = np.pi/(n+1.) # factor
  for i in range(n):
    fac = ((n-i+1)*np.cos(pn*i)+np.sin(pn*i)/np.tan(pn))/(n+1)
    mo[i] *= fac
  return mo



def lorentz_kernel(mus):
  """ Modify coeficient using the Jackson Kernel"""
  mo = mus.copy() # copy array
  n = len(mo)
  pn = np.pi/(n+1.) # factor
  lamb = 3.
  for i in range(n):
    fac = np.sinh(lamb*(1.-i/n))/np.sinh(lamb)
    mo[i] *= fac
  return mo






def fejer_kernel(mus):
  """Default kernel"""
  n = len(mus)
  mo = mus.copy()
  for i in range(len(mus)):
    mo[i] *= (1.-float(i)/n)
  return mo





import numpy as np

# routines to apply strain to Hamiltonian

def add_strain(h,sr,**kwargs):
    """Take as input a Hamiltonian, return the Hamiltonian
    with applied strain"""
    h.turn_multicell() # multicell Hamiltonian
    if not h.has_eh:
      if not h.has_spin: indg = lambda i: i
      elif h.has_spin: indg = lambda i: i//2
      else: raise
    else: 
      if h.has_spin: indg = lambda i: i//4
      else: raise
    f = strain_mode(sr,**kwargs) # get the function depending on the mode
    def fm(m,r1,r2): # function to modify hopping
        return strain_matrix(m,r1,r2,indg,f)
    h.modify_hamiltonian_matrices(fm,use_geometry=True) # modify matrices


def strain_mode(sr,sd=None,mode="scalar"):
    """Create a different fucntion depending on the strain mode"""
    if mode=="scalar": # just change the strength of hoppings
        return lambda r,dr: sr(r)
    elif mode=="directional": # direction-dependent change
        return lambda r,dr: sr(dr)
    elif mode=="non_uniform": # spatial and direction-dependent change
        return lambda r,dr: sr(r,dr)
    else: raise


# NOTE: strain_mode scalar fails with periodic boundary conditions
# the distance function should be fixed, otherwise the Hamiltonain is
# non Hermitian



def uniaxial_strain(H,d=np.array([1.,0.,0.]),s=0.,**kwargs):
    """Add uniaxial strain"""
    d = np.array(d) ; d = d/np.sqrt(d.dot(d)) # normalize
    def fun(dr): # function to get the strain
        dr2 = dr.dot(dr) # moduli
        if dr2<1e-4: return 1.0 # do nothing
        dr = dr/np.sqrt(dr2) # normalize
        dr0 = dr.dot(d) # projection on the direction
        fac = np.abs(dr0) # modulation of the strain
        return 1. + 2*(fac - 0.5)*s # change the hopping
    add_strain(H,fun,mode="directional") # add strain



def strain_matrix(m,rs1,rs2,indg,fs): # function to apply strain to a matrix
    """Apply strain to a matrix, this only works well in 1d"""
    mo = m.copy() # copy matrix
    from scipy.sparse import coo_matrix,csc_matrix
    mo = coo_matrix(mo) # turn sparse
    for k in range(len(mo.data)):
        ii = indg(mo.col[k])
        jj = indg(mo.row[k])
        r0 = (rs1[ii] + rs2[jj])/2. # average location 
        dr0 = rs1[ii] - rs2[jj] # distance difference
        mo.data[k] = fs(r0,dr0)*mo.data[k]
    from .algebra import issparse
    if issparse(m): return csc_matrix(mo)
    else: return mo.todense()


def simple_strain_matrix(m,rs1,rs2,indg,fs): # function to apply strain to a matrix
    mo = m.copy() # copy matrix
    for i in range(len(rs1)): # loop over sites
        for j in range(len(rs2)): # loop over sites
            r0 = (rs1[i]+rs2[j])/2.
            fac = fs(r0)
            for ii in indg(i): # generate indexes for site i
                for jj in indg(j): # generate indexes for site j
                    mo[ii,jj] = fac*m[ii,jj] # strain the matrix
    return mo





import numpy as np
from scipy.optimize import minimize,broyden1
import scipy.linalg as lg


def v2U(v,exp=True):
    """
    Transform a vector into a unitary matrix
    """
    n = int(np.sqrt(v.shape[0]+1)) # dimension of the matrix
    M = np.zeros((n,n),dtype=np.complex_) # define the matrix
    ii = 0 # counter
    d = np.zeros(n,dtype=np.complex_) # diagonal
    for i in range(n-1):
        d[i] = v[ii]
        ii += 1 # increase counter
    d[n-1] = -np.sum(d) # ensure the trace is zero
    for i in range(n): M[i,i] = d[i] # put the diagonal elements
    for j in range(n):
        for i in range(j+1,n):
            M[i,j] = v[ii] + 1j*v[ii+1] # store
            M[j,i] = v[ii] - 1j*v[ii+1] # store
            ii += 2 # increase
    if exp: return lg.expm(1j*M) # return unitary matrix
    else: return M # return the Lie algebra matrix




def v2O(v,exp=True):
    """
    Transform a vector into an orthogonal matrix
    """
    n = int(np.sqrt(v.shape[0]+1)) # dimension of the matrix
    M = np.zeros((n,n)) # define the matrix
    ii = 0 # counter
    for j in range(n):
        for i in range(j+1,n):
            M[i,j] = v[ii]  # store
            M[j,i] = -v[ii]  # store
            ii += 1 # increase
    if exp: return lg.expm(M).real # return orthogonal matrix
    else: # return the algebra matrix (for continuus symmetries)
      return M
#      return M/np.sqrt(np.trace(M@np.transpose(M))) # return normalized matrix



class SymGen():
    """
    Class for the symmetry generator
    """
    def __init__(self,m,discrete=True):
        self.real = np.abs(np.max(m.imag))<1e-7 # real generator
        self.discrete = discrete # discrete symmetry
        self.np = m.shape[0]**2-1 # number of free variables
        self.n = m.shape[0] # dimension of the operator
    def get_x0(self):
        return np.random.random(self.np) # return an initial value
    def get_symmetry(self,x):
        """Return the symmetry operator"""
        if self.real: 
            M = v2O(x,exp=self.discrete)
            if self.discrete: return M # return the operator
            else: return lg.expm(M) # return the operator
        else:
            M = v2U(x,exp=self.discrete)
            if self.discrete: return M # return the operator
            else: return lg.expm(1j*M) # return the operator
    def get_operator(self,x):
        """Return the operator used to compute the commutator"""
        if self.real: return v2O(x,exp=self.discrete)
        else: return v2U(x,exp=self.discrete)
    def return_unique(self,mm):
        """Return the independent symmetry operators"""
        ms = [] 
        for m in mm: 
            if m is not None: ms.append(m) # store this matrix
        if self.discrete: return retain_different(ms)
        else: 
            vs = [m.reshape(self.n**2) for m in ms] # to vectors
            vs = retain_independent(vs) # independent vectors
            ms = [v.reshape((self.n,self.n)) for v in vs]
            ms = [m/np.max(np.abs(m)) for m in ms]
            return ms






def commuting_matrices(m):
    """
    Compute the different unitary operators
    that commute with a matrix.
    Written for SU(N) and SO(N)
    """
    m = np.array(m) # convert to array
    real = np.max(np.abs(m.imag))<1e-3 # real matrix
    n = m.shape[0] # size of the matrix
    SG = SymGen(m,discrete=False) # Generator for discrete symmetry
    def fun(v): # function to minimize
        U = SG.get_operator(v) # get the matrix
        A = m@U - U@m # commutator
        out = np.trace(A@np.conjugate(A.T)) # return trace
        return out.real
#    if real: v0 = np.random.random((n**2-n)//2) # initial guess
# this should be fixed
    def getU(): # return one symmetry operator
      v0 = SG.get_x0() # initial guess
      res = minimize(fun,v0) # minimize
      v = res.x
      U = SG.get_symmetry(v)
      U = SG.get_operator(v)
#      print(np.round(U,2),"Error",fun(v))
      if fun(v)>1e-5: return None
      else: return U
    Us = [getU() for i in range(100)] # compute many of them
    Us = SG.return_unique(Us)
#    Us = retain_different(Us)
    print("Found",len(Us),"symmetries")
    for u in Us:
        print(np.round(u,2))


def retain_different(us):
    """
    Retain operators that are different
    """
    out = []
    for u in us: # loop over input matrices
        if u is None: continue
        store = True # assume that it is different
        for o in out: # loop over stored ones
            if np.max(np.abs(u-o))<1e-4: store = False # already have it
        if store: out.append(u) # store this one
    return out



def retain_generators(us):
    """
    Retain the matrices that generate the symmetry algebra
    """
    out = [] # output list



def retain_independent(M):
    """Return linearly independent vectors"""
    from numpy.linalg import matrix_rank
    dim = len(M)
    tol = 1e-4
    LI=[]
    for i in range(dim):
        tmp=[]
        for r in LI:
            tmp.append(r)
        tmp.append(M[i])   
        if matrix_rank(tmp,tol=tol)>len(LI):    
            LI.append(M[i])      
    return LI       






from __future__ import print_function
import kwant



def transport(ht,energy):
  """Define a Kwant heterostructure using an input one"""
  # create kwant objects
  lat = kwant.lattice.square()
  sys = kwant.Builder() 
  # create the two leads
  sym_L = kwant.TranslationalSymmetry((-1, 0))
  sym_R = kwant.TranslationalSymmetry((1, 0))
  lead_L = kwant.Builder(sym_L) # create the lead
  lead_R = kwant.Builder(sym_R) # create the lead
  lead_R[lat(0,0)] = ht.right_intra
  lead_L[lat(0,0)] = ht.left_intra
  lead_R[lat(0,0),lat(1,0)] = ht.right_inter
#  lead_R[lat.neighbors()] = ht.right_inter
  lead_L[lat(0,0),lat(-1,0)] = ht.left_inter
#  lead_L[lat.neighbors()] = ht.left_inter
  # The scattering region will be three sites
  # add the onsites of the leads
  if not ht.block_diagonal: # dense scattering region
    # and create the coupling to the central region
    # create central and scattering regions
    sys[lat(-1,0)] = ht.left_intra
    sys[lat(1,0)] = ht.right_intra
    sys[lat(0,0)] = ht.central_intra
    sys[lat(0,0), lat(-1,0)] = ht.left_coupling
    sys[lat(0,0), lat(1,0)] = ht.right_coupling
    # now attach the two leads
    sys.attach_lead(lead_L, lat(-1, 0))
    sys.attach_lead(lead_R, lat(1, 0))
  else: # block diagonal form
    nb = len(ht.central_intra)
    sys[lat(-1,0)] = ht.left_intra
    sys[lat(nb,0)] = ht.right_intra
    for i in range(nb):
      sys[lat(i,0)] = ht.central_intra[i][i]
    for i in range(nb-1):
      sys[lat(i,0),lat(i+1,0)] = ht.central_intra[i][i+1]
    sys[lat(0,0), lat(-1,0)] = ht.left_coupling
    sys[lat(nb-1,0), lat(nb,0)] = ht.right_coupling
    sys.attach_lead(lead_L, lat(-1, 0))
    sys.attach_lead(lead_R, lat(nb, 0))
  sys = sys.finalized()
#  kwant.plot(sys)
  smatrix = kwant.smatrix(sys, energy)
  return smatrix.transmission(1, 0)




import numpy as np
from scipy.sparse import coo_matrix
from copy import deepcopy

# TODO
# - more efficient discrete optimizr (computing just the energy correction)
# - autoannealing, stopping the iterations once a reasonable GS is reached



class LatticeGas():
    def __init__(self,g,filling=0.5): # geometry
        g.nrep = 1
        self.geometry = g # store geometry
        self.nsites = len(g.r) # number of sites
        self.mu = np.zeros(len(g.r)) # chemical potential
        self.den = np.zeros(len(g.r)) # chemical potential
        self.j = np.array([0]) # interactions
        self.pairs = np.array([[0,0]]) # empty list
        self.set_filling(filling)
    def set_filling(self,filling):
        """Set filling of the system"""
        self.den[:] = 0. # initialize all to zero
        N = int(np.round(self.nsites*filling)) # filled sites
        self.den = random_density(len(self.den),N) # random density
    def add_interaction(self,Jij=None,**kwargs):
        h = self.geometry.get_hamiltonian(has_spin=False,tij=Jij)
        m = coo_matrix(h.get_hk_gen()([0.,0.,0.])) # get onsite matrix
        pairs = np.array([m.row,m.col]).transpose() # convert to array
        self.pairs = np.concatenate([self.pairs,pairs]) # add interaction
        self.j = np.concatenate([self.j,m.data]).real # store
    def get_energy(self):
        return energy_jax(self.mu,self.pairs,self.j,self.den)
    def get_local_energy(self,**kwargs):
        return get_local_energy(self,**kwargs)
    def get_local_mu(self,**kwargs):
        return get_local_mu(self,**kwargs)
    def optimize_energy(self,**kwargs):
        """Optimize the energy"""
#        print(self.den) ; exit()
        fun = lambda x: energy_jax(self.mu,self.pairs,self.j,x)
        x,es = optimize_discrete(fun,self.den,**kwargs) # optimize
        self.den = x # overwrite
        return es
    def get_correlator(self,**kwargs):
        """Return the nearest neighbor correlators"""
        from .statphystk.correlator import get_nnc
        return get_nnc(self.geometry,self.den,**kwargs)
    def copy(self):
        return deepcopy(self)

from numba import jit

def energy_jax(mu,pairs,js,den):
    if len(pairs)==0:
        pairs = np.array([[0,0]])
        js = np.array([0.])
    return energy_jax_jit(mu,pairs,js,den)


@jit(nopython=True)
def energy_jax_jit(mu,pairs,js,den):
    """Compute the energy of the lattice gas model"""
    nump = len(pairs) # number of pairs
    n = len(mu) # number of sites
    etot = 0. # output energy
    for i in range(n): etot = etot + mu[i]*den[i] # chemical potential
    for ip in range(nump):
        ii = pairs[ip][0]
        jj = pairs[ip][1]
        etot = etot + den[ii]*den[jj]*js[ip] # add contribution
    return etot



def optimize_discrete(fun,x0,temp=0.1,ntries=1e5,info=False):
    """Discrete optimization, using a swap method"""
    inds = np.array(range(len(x0))) # indexes
    vals = np.unique(x0) # different values
    if len(vals)!=2: 
        print(vals)
        raise # not implemented
    def swap(x):
        x = x.copy()
        ainds = inds[x==vals[0]] # indexes for first value
        binds = inds[x==vals[1]] # indexes for second value
#        i1 = np.random.randint(0,n) # one random site
#        i2 = np.random.randint(0,n) # one random site
        j1 = np.random.randint(0,len(ainds)) # one random site
        j2 = np.random.randint(0,len(binds)) # one random site
        i1 = ainds[j1]
        i2 = binds[j2]
        x1 = x[i1]
        x2 = x[i2]
        x[i2] = x1 # swap
        x[i1] = x2 # swap
        return x
    def nswap(x,n):
        for ii in range(n):
            x = swap(x)
        return x
    n = len(x0) # number of sites
    nrep = ntries # this many iterations
    xold = x0.copy()
    es = np.zeros(int(nrep)) # storage for energies
    for ii in range(int(nrep)): # this many iterations
        x = nswap(xold,np.random.randint(1,4)) # make 1,2,3 swaps
        eo = fun(xold) # old
        en = fun(x) # new
        if info: print(en)
        if en<=eo: # smaller or same
            es[ii] = en # store new energy
            xold = x # overwrite
        else: 
            fac = np.exp((eo-en)/temp) # acceptance probability
            if np.random.random()<fac: 
                es[ii] = en # store new energy
                xold = x # overwrite
            else: 
                es[ii] = eo # keep old energy
                pass # do nothing
    return xold,es # return old one



def get_local_energy(LG,normalize=False):
    """Return the local energy at each site for the current snapshot"""
    def get(ii): # get for site ii
        LG0 = LG.copy() # make a dummy copy
        pairs0 = [] # empty list
        j0 = [] # empty list
        for (p,j) in zip(LG.pairs,LG.j): # take only those for this site
            if ii==p[0] or ii==p[1]: # site ii is here
                pairs0.append(p)
                j0.append(j)
        mu0 = LG.mu*0. # zeros
        mu0[ii] = LG.mu[ii]
        LG0.pairs = np.array(pairs0)
        LG0.j = np.array(j0)/2.
        LG0.mu = mu0
        enii = LG0.get_energy()
        if normalize: return enii/np.sum(LG0.j)
        else: return enii
    return np.array([get(ii) for ii in range(len(LG.geometry.r))]) # loop over positions



def get_local_mu(LG,normalize=False):
    """Return the local chemical potential"""
    def get(ii): # get for site ii
        LG0 = LG.copy() # make a dummy copy
        pairs0 = [] # empty list
        j0 = [] # empty list
        for (p,j) in zip(LG.pairs,LG.j): # take only those for this site
            if ii==p[0] or ii==p[1]: # site ii is here
                pairs0.append(p)
                j0.append(j)
        mu0 = LG.mu*0. # zeros
        mu0[ii] = LG.mu[ii]
        LG0.pairs = np.array(pairs0)
        LG0.j = np.array(j0)/2.
        LG0.den[ii] = 1.0 # overwrite to return chemical potential
        LG0.mu = mu0
        enii = LG0.get_energy()
        if normalize: return enii/np.sum(LG0.j)
        else: return enii
    return np.array([get(ii) for ii in range(len(LG.geometry.r))]) # loop over positions



def random_density(Ntot,N):
    """Generate an array with N 1's and Ntot-N 0's,
    with the 1's randomly distributed"""
    out = np.zeros(Ntot) # initialize
    inds = [i for i in range(Ntot)] # indexes with 1's
    for ii in range(N): # loop
        jj = np.random.randint(0,len(inds)) # random number from 0 to randint
        ind = inds[jj] # index with 1
        out[ind] = 1.0
        del inds[jj] # remove this one
    return np.array(out)






import numpy as np

def get_local_energy(CS):
    """Return the local energy at each site for the current snapshot"""
    def get(ii): # get for site ii
        CS0 = CS.copy() # make a dummy copy
        pairs0 = [] # empty list
        j0 = [] # empty list
        for (p,j) in zip(CS.pairs,CS.j): # take only those for this site
            if ii==p[0] or ii==p[1]: # site ii is here
                pairs0.append(p)
                j0.append(j)
        b0 = CS.b*0. # zeros
        b0[ii] = CS.b[ii]
        CS0.pairs = np.array(pairs0)
        CS0.j = np.array(j0)/2.
        CS0.b = b0
        enii = CS0.get_energy()
        return enii
    return np.array([get(ii) for ii in range(len(CS.geometry.r))]) # loop over positions




import numpy as np
from . import extract

def delta2dvector(uu,dd,ud):
    """Transform Deltas to dvectors"""
    out = [(dd + uu)/2.,-1j*(dd-uu)/2,ud] # compute the d-vector
    return np.array(out) # return dvectors (three matrices)


def dvector2delta(ds):
    """Transform a certain dvector into deltauu, deltadd and deltaud"""
    ds = np.array(ds)
    delta = [0.,0.,0.]
    delta[0] = ds[0]-1j*ds[1] # up up
    delta[1] = ds[0]+1j*ds[1] # down down
    delta[2] = ds[2] # up down
    return np.array(delta) # return the delta

dvector2deltas = dvector2delta

def dvector2deltas_old(ds):
    """Transform a certain dvector into deltauu, deltadd and deltaud"""
    raise # this must be checked
    # this function has probably a missing sign somewhere
    deltas = [0.,0.,0.]
    deltas[0] = ds[0]+ds[1]
    deltas[1] = -1j*(ds[0]-ds[1]) # this sign might not be ok
    deltas[2] = ds[2]
    return np.array(deltas)


def extract_dvector_from_hamiltonian(h):
    """Return a function that computes the d-vector matrix at a k-point"""
    hk = h.get_hk_gen() # get Bloch Hamiltonian generator
    def f(k): # define function
        m = hk(k) # compute Bloch Hamiltonian
        (uu,dd,ud) = extract.extract_triplet_pairing(m) # pairing matrices
        return delta2dvector(uu,dd,ud) # return d-vector
    return f # return function

def matrix2dvector(m):
    """Return the dvectors from a matrix"""
    (uu,dd,ud) = extract.extract_triplet_pairing(m) # pairing matrices
    return delta2dvector(uu,dd,ud) # return d-vector


def dvector2nonunitarity(m):
    """Given a matrix of dvectors, compute the non-unitarity"""
    out = np.zeros(m.shape,dtype=np.complex_)
    n = m.shape[1]
    for i in range(n): # loop over sites
        for j in range(n): # loop over sites
            d = m[:,i,j]
            out[:,i,j] = 1j*np.cross(np.conjugate(d),d)
    return out.real


def average_hamiltonian_dvector(h,nk=10,
    spatial_sum=True,
    non_unitarity=False
    ):
    """Compute the average d-vector of a Hamiltonian. Optional arguments
       - nk = 10, number of kpoints in each direction
       - spatial_sum = True, return sum over sites
    """
    if not h.has_eh: raise
    f = extract_dvector_from_hamiltonian(h) # function to extract the d-vector
    ks = h.geometry.get_kmesh(nk=nk) # get k-mesh
    out = np.array([f(k) for k in ks]) # compute d-vector matrices
    # redefine in case you want the non-unitarity
    if non_unitarity: out = [dvector2nonunitarity(o) for o in out]
    out = np.abs(out)**2 # square each term
    out = np.mean(out,axis=0) # average over k-points
    out = np.sum(out,axis=1) # sum over rows
    if spatial_sum: out = np.mean(out,axis=1) # sum over columns
    return out # return a vector

def dvector_times_rij_map(h,nrep=4):
    """Compute the dvector times rij"""
    h = h.supercell(nrep) # create a supercell (if needed)
    hi = h.get_hopping_dict()[(0,0,0)]
    dms = matrix2dvector(hi) # get the dvectors
    rs = h.geometry.r[:,0:3] # get coordinates
    ds = np.zeros(rs.shape,dtype=np.complex_) # array with the result
    for i in range(len(rs)):
        for j in range(len(rs)):
            d = np.cross(dms[:,i,j],rs[i]-rs[j])
            ds[i,:] = ds[i,:] + d # add contribution
    m = np.array([rs[:,0],rs[:,1],rs[:,2],ds[:,0],ds[:,1],ds[:,2]]).T.real
    m = np.round(m,5) # round values
    np.savetxt("DxR_MAP.OUT",m) # write in the file

def dvector_times_mij_map(h,nrep=4):
    """Compute the dvector times rij"""
    h = h.supercell(nrep) # create a supercell (if needed)
    hi = h.get_hopping_dict()[(0,0,0)]
    dms = matrix2dvector(hi) # get the dvectors
    rs = h.geometry.r[:,0:3] # get coordinates
    ds = np.zeros(rs.shape,dtype=np.complex_) # array with the result
    for i in range(len(rs)):
        for j in range(len(rs)):
            d = np.cross(dms[:,i,j],rs[i]-rs[j])
            ds[i,:] = ds[i,:] + d # add contribution
    m = np.array([rs[:,0],rs[:,1],rs[:,2],ds[:,0],ds[:,1],ds[:,2]]).T.real
    m = np.round(m,5) # round values
    np.savetxt("DxR_MAP.OUT",m) # write in the file


def dvector_non_unitarity_map(h,nrep=2,**kwargs):
    """Compute a map of the d-vector non-unitarity"""
    ds = dvector_non_unitarity(h,**kwargs) # dvectors
    rs = h.geometry.supercell(nrep).r # supercell positions
    from ..geometry import replicate_array
    ds = replicate_array(h.geometry,ds,nrep=nrep) # replicate
    m = np.array([rs[:,0],rs[:,1],rs[:,2],ds[:,0],ds[:,1],ds[:,2]]).T
    np.savetxt("NON_UNITARITY_MAP.OUT",m) # write in the file


def dvector_non_unitarity(h,nk=10):
    """Compute the non-unitarity"""
    f = extract_dvector_from_hamiltonian(h) # function to extract the d-vector
    ks = h.geometry.get_kmesh(nk=nk) # get k-mesh
    out = np.array([f(k) for k in ks]) # compute d-vector matrices
    # redefine in case you want the non-unitarity
    out = [dvector2nonunitarity(o) for o in out] # non-unitarity
    out = np.mean(out,axis=0) # average over k-points
    ds = np.sum(out,axis=1).T # sum over rows
    return ds




import numpy as np
import scipy.sparse.linalg as lg
import scipy.linalg as lg

def braket_wAw(w,A):
  w = np.matrix(w) # convert to matrix
  return ((w.T).H*A*w.T)[0,0] # expectation value

def current_operator(h):
  """Get the current operator"""
  h = h.get_multicell()
  def fj(k0):
      return derivative(h,k0,order=[1])
  return fj



def gs_current(h,nk=400):
  weighted_current(h,nk=nk)




def fermi_current(h,nk=400,delta=0.5):
  def fun(e):
    return delta/(delta**2+e**2)*2/np.pi
  weighted_current(h,nk=nk,fun=fun)



def weighted_current(h,nk=400,fun=None):
  """Calculate the Ground state current"""
  if fun is None:
    delta = 0.01
    def fun(e): return (-np.tanh(e/delta) + 1.0)/2.0
  jgs = np.zeros(h.intra.shape[0]) # current array
  hkgen = h.get_hk_gen() # generator
  fj = current_operator(h) # current operator
  ks = np.linspace(0.0,1.0,nk,endpoint=False) # k-points
  for k in ks: # loop
    hk = hkgen(k) # Hamiltonian
    (es,ws) = lg.eigh(hk) # diagonalize
    ws = ws.transpose() # transpose
    jk = fj(k) # get the generator
    for (e,w) in zip(es,ws): # loop
      weight = fun(e) # weight
      print(weight)
      d = np.conjugate(w)*ket_Aw(jk,w) # current density
      jgs += d.real*weight # add contribution
#      jgs += (np.abs(w)**2*weight).real # add contribution
  jgs /= nk # normalize
  print("Total current",np.sum(jgs))
  np.savetxt("CURRENT1D.OUT",np.matrix([range(len(jgs)),jgs]).T)




def derivative(h,k,order=None):
  """Calculate the derivative of the Hamiltonian"""
  ## The order parameter is kind of weird now, this must be fixed ##
  h = h.get_multicell() # get multicell Hamiltonian
  if order is None:
#    order = [1 for i in range(h.dimensionality)] # order of the derivative
    order = [1,0,0] # default
  if h.dimensionality == 0: return None
  elif h.dimensionality == 1: # one dimensional
      mout = h.intra*0.0 # initialize
      for t in h.hopping: # loop over matrices
        phi = np.array(t.dir).dot(k) # phase
        pref = (t.dir[0]*1j)**order[0] # prefactor
        tk = pref*t.m * np.exp(1j*np.pi*2.*phi) # k hopping
        mout = mout + tk # add contribution
      return mout
  elif h.dimensionality == 2: # two dimensional
      k = np.array(k) # convert to array
      mout = h.intra*0.0 # initialize
      for t in h.hopping: # loop over matrices
        d = t.dir
        d = np.array(d) # vector director of hopping
        phi = d.dot(k) # phase
        pref1 = (d[0]*1j)**order[0] # prefactor
        pref2 = (d[1]*1j)**order[1] # prefactor
        pref = pref1*pref2 # total prefactor
        tk = pref*t.m * np.exp(1j*np.pi*2.*phi) # derivative of the first
        mout = mout + tk # add to the hamiltonian
      return mout
  else: raise # not implemented







import numpy as np


adaptive_error = 1e-2 # error in adaptive algorithm


def integrate_matrix(f,xlim=[0.,1.],eps=0.1,only_imag=False):
  """ Integrates a matrix, the measure is the maximun value of the matrix"""
  return adaptive_simpsons_rule(f,xlim[0],xlim[1],eps,only_imag=only_imag)



def integrate_matrix_2D(f,xlim=[0.,1.],ylim=[0.,1.],**kwargs):
    """Perform a 2D integral of a matrix"""
    def intx(x): # inegrate in the y axis
        return integrate_matrix(lambda y: f([x,y]),xlim=ylim,**kwargs)
    return integrate_matrix(intx,xlim=xlim,**kwargs) # now integrate in x axis
    



def simpsons_rule(f,a,b):
    c = (a+b) / 2.0
    h3 = abs(b-a) / 6.0
    return h3*(f(a) + 4.0*f(c) + f(b))
 
def recursive_asr(f,a,b,eps,whole):
    "Recursive implementation of adaptive Simpson's rule."
    c = (a+b) / 2.0
    left = simpsons_rule(f,a,c)
    right = simpsons_rule(f,c,b)
    if np.max(np.abs(left + right - whole)) <= 15*eps:
        return left + right + (left + right - whole)/15.0
    return recursive_asr(f,a,c,eps/2.0,left) + recursive_asr(f,c,b,eps/2.0,right)


def recursive_asr_imag(f,a,b,eps,whole):
    "Recursive implementation of adaptive Simpson's rule."
    c = (a+b) / 2.0
    left = simpsons_rule(f,a,c)
    right = simpsons_rule(f,c,b)
    if np.max(np.abs(np.imag(left + right - whole))) <= 15*eps:
        return left + right + (left + right - whole)/15.0
    return recursive_asr(f,a,c,eps/2.0,left) + recursive_asr(f,c,b,eps/2.0,right)



def simpson(f,xlim=[0.,1.],eps=0.1):
  """ Integrates a matrix, the measure is the maximun value of the matrix"""
  a = xlim[0]
  b = xlim[1]
  return scalar_asr(f,a,b,eps,simpsons_rule(f,a,b))



def scalar_asr(f,a,b,eps,whole):
    "Recursive implementation of adaptive Simpson's rule."
    c = (a+b) / 2.0
    left = simpsons_rule(f,a,c)
    right = simpsons_rule(f,c,b)
    if np.abs(left + right - whole) <= 15*eps:
        return left + right + (left + right - whole)/15.0
    return scalar_asr(f,a,c,eps/2.0,left) + scalar_asr(f,c,b,eps/2.0,right)



 
def adaptive_simpsons_rule(f,a,b,eps,only_imag=False):
    "Calculate integral of f from a to b with max error of eps."
    if only_imag:
      return recursive_asr_imag(f,a,b,eps,simpsons_rule(f,a,b))
    else:
      return recursive_asr(f,a,b,eps,simpsons_rule(f,a,b))



def random_integrate(f,df=1e-4):
    """Iterate a function until its result shows small enough fluctuations"""
    y0 = f() # compute once
    yt = y0 + 0. # initialize the total result
    yold = y0 + 0. # initialize the old result
    it = 0 # initialize
    while True: # infinite loop
      it += 1 # number of iterations
      y0 = f() # execute the function
      yt += y0 # add contribution
      diff = yold/it - yt/(it+1.) # difference between current and previous
      if np.mean(np.abs(diff))<np.mean(np.max(yt/(it+1.)))*df:
          break
      yold = yt + 0. # update
    return yt/(it+1.) # return result



def peak_integrate(f,x0,x1,xp=0.0,dp=1e-6,**kwargs):
    """Wrapper to quad, for function that has a peak"""
    from scipy.integrate import quad
    i0 = quad(f,x0,xp-dp,**kwargs) # first interval
    i1 = quad(f,xp-dp,xp+dp,**kwargs) # second interval
    i2 = quad(f,xp+dp,x1,**kwargs) # third interval
    return i1+i2+i2
  



def complex_contour(f,xmin=-5,xmax=0,eps=1e-2,mode="upper",
                              **kwargs):
    """Perform an integral using a complex contour"""
#    return integrate_matrix(f,xlim=[xmin,xmax],**kwargs) # integrate matrix
    def fint(x):
        if mode=="upper": 
            z0 = (xmin-xmax)*np.exp(-1j*x*np.pi)/2.
            z = z0 + (xmin+xmax)/2.
            return -1j*(f(z)*z0)*np.pi # integral after change of variables
        elif mode=="lower": 
            z0 = (xmin-xmax)*np.exp(1j*x*np.pi)/2.
            z = z0 + (xmin+xmax)/2.
            return 1j*(f(z)*z0)*np.pi # integral after change of variables
    y = fint(xmin) # evaluate
#    if type(y)=np.array:
    return integrate_matrix(fint,xlim=[0.,1.],eps=eps,**kwargs) # integrate matrix
#    else:
#        import scipy.integrate as integrate
#        return integrate.quad(fint2,0.,1.0,limit=60,epsabs=0.1,epsrel=0.1)[0]
        








import numpy as np

# routines to compute the dIdV using the embedding method

def get_didv(self,T=1e-2,write=True,nsuper=1,**kwargs):
    from ..transporttk.localprobe import LocalProbe
    lp = LocalProbe(self,T=T,**kwargs) # local probe object
    lp.reuse_gf = True # reuse the Green's function
    # now we will overwrite a few objects
    # this is not very elegant, but it works
    g = lp.H.geometry.get_supercell(nsuper) # supercell geometry
    Hc = lp.H.copy() # copy the original object
    # for the selfenernergy, the intracell is picked from lp.H
    lp.H = self.H.get_supercell(nsuper) # overwrite Hamiltonian (for the intra)
    # the Green's function is now directly computed for the supercell
    lp.H.get_gf = lambda **kwargs: Hc.get_gf(nsuper=nsuper,**kwargs)
    # now that the methods are overwritten, lets compute
    Gs = [] # conductances
    for i in range(len(g.r)): # loop over positions
        lp.i = i # update position
        Gs.append(lp.didv(**kwargs)) # compute this site
    if write:
        np.savetxt("DIDV.OUT",np.array([g.r[:,0],g.r[:,1],np.array(Gs)]).T)
    return g.r[:,0],g.r[:,1],np.array(Gs)





import numpy as np


def hamiltonian_fusion(h1,h2):
    """Given two Hamiltonians, fusion them"""
    go = h1.geometry + h2.geometry # sum the geometries
    h1 = h1.get_multicell() # to multicell form
    ho = h1.copy() # copy Hamiltonian
    ho.geometry = go # update geometry
    d1 = h1.get_dict() # get dictionary
    d2 = h2.get_dict() # get dictionary
    from ..multihopping import direct_sum_hopping_dict
    do = direct_sum_hopping_dict(d1,d2)
    from ..multihopping import MultiHopping
    mho = MultiHopping(do)
    ho.set_multihopping(mho)
    return ho # return Hamiltonian






from __future__ import print_function
import scipy.sparse.linalg as slg
import scipy.linalg as lg
from scipy.sparse import csc_matrix as csc
from scipy.sparse import csc_matrix 
from scipy.sparse import bmat
import os
from numba import jit
import numpy as np
from . import klist
from . import operators
from . import timing
from . import parallel
from . import algebra
from .increase_hilbert import full2profile as spatial_dos
from . import filesystem as fs

def ldos0d(h,e=0.0,delta=0.01,write=True):
  """Calculates the local density of states of a Hamiltonian and
     writes it in file"""
  if h.dimensionality==0:  # only for 0d
    iden = np.identity(h.intra.shape[0],dtype=np.complex_) # create identity
    g = ( (e+1j*delta)*iden -h.intra ).I # calculate green function
  else: raise # not implemented...
  d = [ -(g[i,i]).imag/np.pi for i in range(len(g))] # get imaginary part
  d = spatial_dos(h,d) # convert to spatial resolved DOS
  g = h.geometry  # store geometry
  if write: write_ldos(g.x,g.y,d,z=g.z) # write in file
  return d



def dos_site_kpm(h,energies=np.linspace(-1.,1.,1000),
        delta=0.01,scale = 10.0,i=0,nk=5,sector=None):
    """Compute a local DOS using the KPM"""
    h = h.copy()
    h.turn_sparse()
    npol = 5*int(scale/delta) # number of polynomials
    from . import kpm
    def getm(m,j): # for the matrix
      es,ds = kpm.ldos(m,i=j,scale=scale,npol=npol,ne=npol*5)
      return es,ds # return
    def get(j): # for the kpoint
        ks = klist.kmesh(h.dimensionality,nk=nk)
        hk = h.get_hk_gen() # get generator
        out = [getm(hk(k),j) for k in ks] # compute all
        ds = np.mean([o[1] for o in out],axis=0)
        es = out[0][0]
        return es,ds
#        if h.dimensonality!=0: raise # not implemented
#        return getm(h.intra)
    from scipy.interpolate import interp1d
    if h.has_spin and h.has_eh: # spin with electron-hole
        if sector is None:
            (es,ds1) = get(4*i)
            (es,ds2) = get(4*i+1)
            (es,ds3) = get(4*i+2)
            (es,ds4) = get(4*i+3)
            ds = ds1+ds2+ds3+ds4
        elif sector=="electron":
            (es,ds1) = get(4*i)
            (es,ds2) = get(4*i+1)
            ds = ds1+ds2
        else: raise
    elif h.has_spin and not h.has_eh: # spinful
        (es,ds1) = get(2*i)
        (es,ds2) = get(2*i+1)
        ds = ds1+ds2
    elif not h.has_spin and not h.has_eh: # spinless
        (es,ds) = get(i)
    else: raise
    f = interp1d(es,ds.real,bounds_error=False,fill_value=0.0)
    return energies,f(energies)





def dos_site(h,i=0,mode="ED",energies=np.linspace(-1.,1.,500),**kwargs):
    """DOS in a particular site for different energies"""
    if mode=="ED":
      if h.dimensionality!=0: raise # only for 0d
      out = []
      for e in energies:
          d = ldos0d(h,e=e,write=False,**kwargs)
          out.append(d[i]) # store
      return (energies,np.array(out)) # return result
    elif mode=="KPM":
        return dos_site_kpm(h,energies=energies,i=i,**kwargs)






def ldos0d_wf(h,e=0.0,delta=0.01,num_wf = 10,robust=False,tol=0):
  """Calculates the local density of states of a hamiltonian and
     writes it in file, using arpack"""
  if h.dimensionality==0:  # only for 0d
    intra = csc_matrix(h.intra) # matrix
  else: raise # not implemented...
  if robust: # go to the imaginary axis for stability
    eig,eigvec = slg.eigs(intra,k=int(num_wf),which="LM",
                        sigma=e+1j*delta,tol=tol) 
    eig = eig.real # real part only
  else: # Hermitic Hamiltonian
    eig,eigvec = slg.eigsh(intra,k=int(num_wf),which="LM",sigma=e,tol=tol) 
  d = np.array([0.0 for i in range(intra.shape[0])]) # initialize
  for (v,ie) in zip(eigvec.transpose(),eig): # loop over wavefunctions
    v2 = (np.conjugate(v)*v).real # square of wavefunction
    fac = delta/((e-ie)**2 + delta**2) # factor to create a delta
    d += fac*v2 # add contribution
#  d /= num_wf # normalize
  d /= np.pi # normalize
  d = spatial_dos(h,d) # resum if necessary
  g = h.geometry  # store geometry
  write_ldos(g.x,g.y,d,z=g.z) # write in file




def ldos_arpack(intra,num_wf=10,robust=False,tol=0,e=0.0,delta=0.01):
  """Use arpack to calculate hte local density of states at a certain energy"""
  if robust: # go to the imaginary axis for stability
    eig,eigvec = slg.eigs(intra,k=int(num_wf),which="LM",
                        sigma=e+1j*delta,tol=tol) 
    eig = eig.real # real part only
  else: # Hermitic Hamiltonian
    eig,eigvec = slg.eigsh(intra,k=int(num_wf),which="LM",sigma=e,tol=tol) 
  d = np.array([0.0 for i in range(intra.shape[0])]) # initialize
  for (v,ie) in zip(eigvec.transpose(),eig): # loop over wavefunctions
    v2 = (np.conjugate(v)*v).real # square of wavefunction
    fac = delta/((e-ie)**2 + delta**2) # factor to create a delta
    d += fac*v2 # add contribution
#  d /= num_wf # normalize
  d /= np.pi # normalize
  return d



def ldos_waves(intra,es = [0.0],delta=0.01,operator=None,
        num_bands=None,k=None,delta_discard=None,**kwargs):
  """Calculate the DOS in a set of energies by full diagonalization"""
  es = np.array(es) # array with energies
  if num_bands is None:
      eig,eigvec = algebra.eigh(intra) 
  else:
      eig,eigvec = algebra.smalleig(intra,numw=num_bands,evecs=True)
      eigvec = eigvec.T
  ds = [] # empty list
  if operator is None: weights = eig*0. + 1.0
  else: weights = [operator.braket(v,k=k) for v in eigvec.transpose()] # weights
  if delta_discard is not None: # discard too far values
      ewin = [min(es)-delta_discard*delta,min(es)+delta_discard*delta]
      for i in range(len(weights)):
          e = eig[i]
          if not ewin[0]<e<ewin[1]: weights[i] = 0.0
  v2s = [(np.conjugate(v)*v).real for v in eigvec.transpose()]
  ds = [[0.0 for i in range(intra.shape[0])] for e in es] # initialize
  ds = ldos_waves_jit(np.array(es),
          np.array(eigvec).T,np.array(eig),np.array(weights),
          np.array(v2s),np.array(ds),delta)
  return ds



@jit(nopython=True)
def ldos_waves_jit(es,eigvec,eig,weights,v2s,ds,delta):
  for i in range(len(es)): # loop over energies
    energy = es[i] # energy
    d = ds[i]
    for j in range(len(eig)):
        v = eigvec[j]
        ie = eig[j]
        weight = weights[j]
        v2 = v2s[j]
        fac = delta/((energy-ie)**2 + delta**2) # factor to create a delta
        d += weight*fac*v2 # add contribution
    d /= np.pi # normalize
    ds[i] = d # store
  return ds


def ldos_diagonalization(m,e=0.0,**kwargs):
    """Compute the LDOS using exact diagonalization"""
#    if algebra.issparse(m): return ldos_arpack(m,e=e,**kwargs) # sparse
    return ldos_waves(m,es=[e],**kwargs)[0] # dense




def ldosmap(h,energies=np.linspace(-1.0,1.0,40),delta=None,
        nk=40,operator=None,**kwargs):
  """Write a map of the ldos using full diagonalization"""
  if h.dimensionality == 0: nk = 1
  if delta is None:
    delta = (np.max(energies)-np.min(energies))/len(energies) # delta
  hkgen = h.get_hk_gen() # get generator
  dstot = np.zeros((len(energies),h.intra.shape[0])) # initialize
  operator = h.get_operator(operator)
  def getd(k): # get LDOS
    hk = hkgen(k) # get Hamiltonian
    # LDOS for this kpoint
    ds = ldos_waves(hk,k=k,es=energies,delta=delta,operator=operator,**kwargs) 
    return ds
  ks = [np.random.random(3) for ik in range(nk)] # kpoints
  ds = parallel.pcall(getd,ks) # get densities
  dstot = np.mean(ds,axis=0) # average over first axis
  print("LDOS finished")
  dstot = [spatial_dos(h,d) for d in dstot] # convert to spatial resolved DOS
  return energies,np.array(dstot)



def spatial_energy_profile(h,**kwargs):
  """Computes the DOS for each site of an slab, only for 2d"""
  if h.dimensionality==0:
      pos = h.geometry.r[:,0]
  elif h.dimensionality==1: pos = h.geometry.r[:,1]
  elif h.dimensionality==2: pos = h.geometry.r[:,2]
  else: raise
  es,ds = ldosmap(h,**kwargs)
  if len(ds[0])!=len(pos): 
    print("Wrong dimensions",len(ds[0]),len(pos))
    raise
  f = open("DOSMAP.OUT","w")
  f.write("# energy, index, DOS, position\n")
  for ie in range(len(es)):
    for ip in range(len(pos)):
      f.write(str(es[ie])+"  ")
      f.write(str(ip)+"  ")
      f.write(str(ds[ie,ip])+"   ")
      f.write(str(pos[ip])+"\n")
  f.close()
  return es,np.transpose(ds) # retunr LDOS 



slabldos = spatial_energy_profile # redefine





def ldos1d(h,e=0.0,delta=0.001,nrep=3):
  """ Calculate DOS for a 1d system"""
  from . import green
  if h.dimensionality!=1: raise # only for 1d
  gb,gs = green.green_renormalization(h.intra,h.inter,energy=e,delta=delta)
  d = [ -(gb[i,i]).imag for i in range(len(gb))] # get imaginary part
  d = spatial_dos(h,d) # convert to spatial resolved DOS
  g = h.geometry  # store geometry
  x,y = g.x,g.y # get the coordinates
  go = h.geometry.copy() # copy geometry
  go = go.supercell(nrep) # create supercell
  write_ldos(go.x,go.y,d.tolist()*nrep) # write in file
  return d


def ldos_projector(h,e=0.0,**kwargs):
    """Return an operator to project onto that region"""
    (x,y,d) = ldos(h,e=e,mode="arpack",silent=True,write=False,**kwargs)
    inds = np.array(range(len(d))) # indexes
    n = len(d)
    d = d/np.sum(d) # normalize
    m = csc_matrix((d,(inds,inds)),shape=(n,n),dtype=np.complex_) # matrix
    m = h.spinless2full(m) # to full matrix
    return operators.Operator(m) # convert to operator

def ldos_density(h,**kwargs):
    """Return a normalized profile with the DOS"""
    (x,y,d) = ldos(h,mode="arpack",silent=True,write=False,**kwargs)
    d = d/np.sum(d) # normalize
    return d


def ldos_potential(h,**kwargs):
    """Return a function that evaluates an LDOS profile"""
    return # not finished yet




def get_ldos(h,projection="TB",**kwargs):
    """ Calculate LDOS"""
    if projection=="TB": return get_ldos_tb(h,**kwargs)
    elif projection=="TBRS": return get_ldos_tb(h,interpolate=True,**kwargs)
    elif projection=="atomic": 
        from .ldostk import atomicmultildos
        return atomicmultildos.get_ldos(h,**kwargs)
    else: raise


def get_ldos_tb(h,e=0.0,delta=0.001,nrep=5,nk=None,ks=None,mode="arpack",
             random=False,silent=True,interpolate=False,
             operator=None,return_rd = False,
             write=True,**kwargs):
    """ Calculate LDOS in a tight binding basis"""
    if ks is not None and mode=="green": raise
    if operator is not None: operator = h.get_operator(operator)
    if mode=="green":
      from . import green
      if h.dimensionality!=2: raise # only for 2d
      h = h.copy()
      h.turn_dense()
      if nk is not None:
        print("LDOS using normal integration with nkpoints",nk)
        gb,gs = green.bloch_selfenergy(h,energy=e,delta=delta,mode="full",nk=nk)
        d = [ -(gb[i,i]).imag for i in range(len(gb))] # get imaginary part
      else:
        print("LDOS using renormalization adaptative Green function")
        gb,gs = green.bloch_selfenergy(h,energy=e,delta=delta,mode="adaptive")
        d = [ -(gb[i,i]).imag/np.pi for i in range(len(gb))] # get imaginary part
    elif mode=="arpack" or mode=="diagonalization": # arpack diagonalization
      from . import klist
      if nk is None: nk = 10
      hkgen = h.get_hk_gen() # get generator
      ds = [] # empty list
      if ks is None:
        ks = klist.kmesh(h.dimensionality,nk=nk)
        if random: ks = [np.random.random(3) for k in ks] # random mesh
      ts = timing.Testimator(title="LDOS",maxite=len(ks),silent=silent)
      for k in ks: # loop over kpoints
        ts.iterate()
        hk = hkgen(k) # get Hamiltonian
        ds += [ldos_diagonalization(hk,e=e,delta=delta,**kwargs)]
      d = np.mean(ds,axis=0) # average
    else: raise # not recognized
    # write result
    d = spatial_dos(h,d) # convert to spatial resolved DOS
    g = h.geometry  # store geometry
    x,y,z,r = g.x,g.y,g.z,g.r # get the coordinates
    if nrep>1:
        go = h.geometry.copy() # copy geometry
        go = go.supercell(nrep) # create supercell
        do = d.tolist()*(nrep**g.dimensionality) # replicate
        xo = go.x
        yo = go.y
        zo = go.z
        ro = go.r
    else: xo,yo,zo,ro,do = x,y,z,r,d
    if interpolate:
        from .interpolation import atomic_interpolation
        xo,yo,do = atomic_interpolation(xo,yo,do,**kwargs)
        if return_rd: raise # not implemented
    if write: 
#        if return_rd: raise # not implemented
        write_ldos(xo,yo,do) # write in file
    xo = np.array(xo)
    yo = np.array(yo)
    zo = np.array(zo)
    ro = np.array(ro)
    do = np.array(do)
    if return_rd: # return positions and DOS
        return (ro,do) # return LDOS
    else: # return x,y and DOS
        return (xo,yo,do) # return LDOS


ldos = get_ldos # for backcompatibility

def multi_ldos(h,projection="TB",**kwargs):
    """Compute the LDOS at different energies, and save everything in a file"""
    if projection=="TB": return multi_ldos_tb(h,**kwargs)
    elif projection=="atomic": 
        from .ldostk import atomicmultildos
        return atomicmultildos.multi_ldos(h,**kwargs)


def multi_ldos_tb(h,es=np.linspace(-1.0,1.0,100),delta=0.01,
        nrep=3,nk=100,num_bands=20,
        random=False,op=None,**kwargs):
  """Calculate many LDOS, by diagonalizing the Hamiltonian"""
  print("Calculating eigenvectors in LDOS")
  ps = [] # weights
  evals,ws = [],[] # empty list
  ks = klist.kmesh(h.dimensionality,nk=nk) # get grid
  hk = h.get_hk_gen() # get generator
  op = operators.tofunction(op) # turn into a function
#  if op is None: op = lambda x,k: 1.0 # dummy function
  if h.is_sparse: # sparse Hamiltonian
    from .bandstructure import smalleig
    print("SPARSE Matrix")
    for k in ks: # loop
      print("Diagonalizing in LDOS, SPARSE mode")
      if random:
        k = np.random.random(3) # random vector
        print("RANDOM vector in LDOS")
      e,w = smalleig(hk(k),numw=num_bands,evecs=True)
      evals += [ie for ie in e]
      ws += [iw for iw in w]
      ps += [op(iw,k=k) for iw in w] # weights
  else:
    print("Diagonalizing in LDOS, DENSE mode")
    for k in ks: # loop
      if random:
        k = np.random.random(3) # random vector
        print("RANDOM vector in LDOS")
      e,w = algebra.eigh(hk(k))
      w = w.transpose()
      evals += [ie for ie in e]
      ws += [iw for iw in w]
      ps += [op(iw,k=k[0]) for iw in w] # weights
#      evals = np.concatenate([evals,e]) # store
#      ws = np.concatenate([ws,w]) # store
  ds = [(np.conjugate(v)*v).real for v in ws] # calculate densities
  del ws # remove the wavefunctions
  fs.rmdir("MULTILDOS") # remove folder
  fs.mkdir("MULTILDOS") # create folder
  go = h.geometry.copy() # copy geometry
  go = go.supercell(nrep) # create supercell
  fo = open("MULTILDOS/MULTILDOS.TXT","w") # files with the names
  def getldosi(e):
    """Get this iteration"""
    out = np.array([0.0 for i in range(h.intra.shape[0])]) # initialize
    for (d,p,ie) in zip(ds,ps,evals): # loop over wavefunctions
      fac = delta/((e-ie)**2 + delta**2) # factor to create a delta
      out += fac*d*p # add contribution
    out /= np.pi # normalize
    return spatial_dos(h,out) # resum if necessary
  outs = parallel.pcall(getldosi,es) # get energies
  ie = 0
  for e in es: # loop over energies
    print("MULTILDOS for energy",e)
    out = outs[ie] ; ie += 1 # get and increase
    name0 = "LDOS_"+str(e)+"_.OUT" # name of the output
    name = "MULTILDOS/" + name0
    from .geometry import replicate_array
    write_ldos(go.x,go.y,replicate_array(h.geometry,out,nrep=nrep),
                  output_file=name) # write in file
    fo.write(name0+"\n") # name of the file
    fo.flush() # flush
  fo.close() # close file
  fmap = open("DOSMAP.OUT","w")
  for ii in range(len(h.geometry.x)):
      for ie in range(len(es)):
          fmap.write(str(ii)+"  ")
          fmap.write(str(es[ie])+"  ")
          fmap.write(str(outs[ie][ii])+"\n")
  fmap.close()
  # Now calculate the DOS
  from .dos import calculate_dos
  es2 = np.linspace(min(es),max(es),len(es)*10)
  ys = calculate_dos(evals,es2,delta,w=None) # compute DOS
  from .dos import write_dos
  write_dos(es2,ys,output_file="MULTILDOS/DOS.OUT")  





def write_ldos(x,y,dos,output_file="LDOS.OUT",z=None):
  """ Write LDOS in a file"""
  fd = open(output_file,"w")   # open file
  fd.write("# x,  y, local density of states\n")
  ii = 0
  for (ix,iy,idos) in zip(x,y,dos): # write everything
    fd.write(str(ix) +"   "+ str(iy) + "   "+ str(idos))
    if z is not None: fd.write("   "+str(z[ii]))
    fd.write("\n")
    ii += 1
  fd.close() # close file




def ldos_finite(h,e=0.0,n=10,nwf=4,delta=0.0001):
  """Calculate the density of states for a finite system"""
  if h.dimensionality!=1: raise # if it is not one dimensional
  intra = csc(h.intra) # convert to sparse
  inter = csc(h.inter) # convert to sparse
  interH = inter.H # hermitian
  m = [[None for i in range(n)] for j in range(n)] # full matrix
  for i in range(n): # add intracell
    m[i][i] = intra
  for i in range(n-1): # add intercell
    m[i][i+1] = inter
    m[i+1][i] = interH
  m = bmat(m) # convert to matrix
  (ene,wfs) = slg.eigsh(m,k=nwf,which="LM",sigma=0.0) # diagonalize
  wfs = wfs.transpose() # transpose wavefunctions
  dos = (wfs[0].real)*0.0 # calculate dos
  for (ie,f) in zip(ene,wfs): # loop over waves
    c = 1./(1.+((ie-e)/delta)**2) # calculate coefficient
    dos += np.abs(f)*c # add contribution
  odos = spatial_dos(h,dos) # get the spatial distribution
  go = h.geometry.supercell(n) # get the supercell
  write_ldos(go.x,go.y,odos) # write in a file
  return dos # return the dos





def ldos_defect(h,v,e=0.0,delta=0.001,n=1):
  """Calculates the LDOS of a cell with a defect, writting the n
  neighring cells"""
  raise # still not finished
  from . import green
  # number of repetitions
  rep = 2*n +1
  # calculate pristine green function
  g,selfe = green.supercell_selfenergy(h,e=e,delta=delta,nk=100,nsuper=rep)
  # now calculate defective green function 
  ez = e + 1j*delta # complex energy
  emat = np.matrix(np.identity(len(g)))*ez  # E +i\delta 
  from . import supercell
  pintra = supercell.intra_super2d(h,n=rep) # pristine
  vintra = supercell.intra_super2d(h,n=rep,central=v) # defective
  selfe = emat - pintra - g.I # dyson euqation, get selfenergy
  gv = (emat - vintra -selfe).I   # Green function of a vacancy, with selfener
  return



from .ldostk.ldosr import ldosr_generator





import numpy as np
from .. import algebra
from numba import jit

def rkky(h,R=[0,0,0],ii=0,jj=0,**kwargs):
    get = rkky_generator(h,**kwargs) # get generator
    return get(R,ii,jj) # evaluate once


def rkky_generator(h,delta=None,nk=100):
    """Compute the RKKY interaction for the ground state
    - R location of the unit cell
    - delta analytic continuation
    """
    # R should be smaller than nk
    h = h.copy()
    h.remove_spin() # remove spin degree of freedom
    es,ws,ks = h.get_eigenvectors(kpoints=True,nk=nk) # get eigenvectors
    if delta is None: delta = 1./nk
    def get(R,ii,jj):
        d1s = np.array([w[ii] for w in ws]) # densities
        d2s = np.array([w[jj] for w in ws]) # densities
        R = np.array(R) # convert to array
        phis = np.array([h.geometry.bloch_phase(R,k) for k in ks]) # phase
        fs = (np.tanh(es/delta) + 1.0)/2. # smearing
        nw = len(es)/len(h.geometry.r)
        return rkky_loop(es,phis,fs,d1s,d2s,delta)/(nw**2) # return RKKY int
    return get



@jit(nopython=True)
def rkky_loop(es,phis,fs,d1s,d2s,delta):
    """Summation for the RKKY interaction"""
    n = len(es) # number of states
    out = 0.0j # output
    delta2 = delta**2
    for i in range(n): # loop
        ei,phii,fi = es[i],phis[i],fs[i]
        for j in range(n): # loop
            ej,phij,fj = es[j],phis[j],fs[j]
            phi = phii/phij # relative phase
            dd = d1s[i]*np.conjugate(d1s[j])
            dd *= np.conjugate(d2s[i])*d2s[j]
            out += phi*1./(ei-ej+1j*delta2)*(fi-fj)*dd
    return out.real


def rkky_pm(h,nk=20,delta=1e-1,**kwargs):
    """Compute the RKKY using a poor man convolution of the k-DOS"""
    raise
    if h.dimensionality!=2: raise
    qs0 = h.geometry.get_kmesh(nk=nk)
    qs = np.array([fR(q) for q in qs0]) # convert
    from ..fermisurface import fermi_surface_generator
    energies = [0.0] # above and below Fermi
    es,ks,ds = fermi_surface_generator(h,reciprocal=False,info=info,
            energies=energies,delta=delta,
            nk=nk,**kwargs)
    # we now have the energies, k-points and DOS, lets do a convolution
    d = poor_man_qpi_single_energy(ks,ds[:,i],qs) # parallel function
    out = parallel.pcall(fp,range(len(es))) # compute in parallel






import numpy as np

def multireplicas(self,n):
    """
    Return several replicas of the unit cell, it is similar
    to supercell but without the shift of the center and going to positive
    and negative positions
    """
    if n==0: return self.r # return positions
    out = [] # empty list with positions
    dl = self.neighbor_directions(n) # list with neighboring cells to take
    if self.dimensionality==0: return self.r
    else: # bigger dimensionality
        for d in dl:  
            out += self.replicas(d=d).tolist() # add this direction
    return np.array(out)

def replicas(self,d=[0.,0.,0.]):
    """Return replicas of the atoms in the unit cell"""
    return np.array([ri + self.a1*d[0] + self.a2*d[1] + self.a3*d[2] for ri in self.r])


def replicate_site(self,ri,n=2):
    """Make replicas of a single site"""
    if n==0: return [self.ri] # return positions
    out = [] # empty list with positions
    dl = self.neighbor_directions(n) # list with neighboring cells to take
    if self.dimensionality==0: return [self.ri]
    else: # bigger dimensionality
        out = np.array([ri + self.a1*d[0] + self.a2*d[1] + self.a3*d[2] for d in dl])
    return np.array(out) # return positions



def get_distance2closest(r,g=None):
    """Distance to the closest site, including replicas"""
    if g is None: rs = [r]
    else: rs = replicate_site(g,r) # return all the replicas
    def fdis(r,r0): # function to compute the distance
        do = np.zeros(len(rs)) # initialize
        for i in range(len(rs)):
            dr = rs[i] - r0 # distance
            do[i] = np.sqrt(dr.dot(dr)) # store
        return np.min(do) # minimum
    return fdis # return function










from . import fullgreen
import numpy as np
from ..increase_hilbert import full2profile as spatial_dos

def ldos(ht,operator=None,write=True,nsuper=None,kpath=None,**kwargs):
    """Compute the local density of states"""
    def get(ht):
        if ht.dimensionality!=1: raise
        if ht.block_diagonal:
            nc = len(ht.central_intra) # number of cells in the middle
            ls = [] # storage for LDOS
            if operator is not None:
                op = ht.Hr.get_operator(operator) # redefine operator
            for ic in range(nc): # loop
                g = fullgreen.get_full_green(ht,mode="central",ic=ic,**kwargs)
                if operator is not None: g = op*g # redefine
                d = [ -(g[i,i]).imag/np.pi for i in range(len(g))] # LDOS
                d = spatial_dos(ht.Hr,d) # convert to spatial resolved DOS
                ls = np.concatenate([ls,d]) # store LDOS
        else: raise
        return nc,ls # return positions and LDOS
    if ht.dimensionality==1: 
        nc,ls = get(ht) # 1d
        r = ht.Hr.geometry.get_supercell(nc).r # generate supercell
    elif ht.dimensionality==2: 
        if kpath is None: kpath = np.linspace(0.,1.,200)
        from ..parallel import pcall
        out = pcall(lambda k: get(ht.generate(k)),kpath) # parallelize over k
        nc = out[0][0] # positions
        ls = 0.
        for i in range(len(out)): ls = ls + out[i][1]
        ls = ls/len(kpath) # normalize
        # now generate the positions
        g = ht.Hr.geometry # original geometry
        from .. import sculpt
        g = sculpt.rotate_a2b(g,g.a2,np.array([1.0,0.0,0.0]))
        gs = g.get_supercell([nc,1]) # supercell in junction dir
        if nsuper is None: nsuper = nc
        r = gs.get_supercell([1,nsuper]).r # supercell in invariant dir
        # now replicate the LDOS
        lso = []
        for i in range(nsuper): lso = np.concatenate([lso,ls])
        ls = lso # overwrite
    else: raise
    if write:
        np.savetxt("LDOS.OUT",np.array([r[:,0],r[:,1],ls]).T)
    return r[:,0],r[:,1],ls








import numpy as np
from numba import jit

import scipy.linalg as lg

#use_fortran = False
try:
    from . import dyson2df90
    use_fortran = True
except:
    use_fortran = False

use_fortran = False # do not use the Fortran routines

def dyson(h,nsuper,nk,ez):
    if h.is_multicell: h = h.get_no_multicell()
    h.turn_dense() # dense matrices
    if h.dimensionality==0: raise
    elif h.dimensionality==1: 
        return dyson1d(h.intra,h.inter,nsuper[0],nk,ez)
    elif h.dimensionality==2: 
        return dyson2d(h.intra,h.tx,h.ty,h.txy,h.txmy,nsuper[0],nsuper[1],nk,ez)
    else: raise




def dyson2d(intra,tx,ty,txy,txmy,nx,ny,nk,ez):
    if use_fortran:
        from . import dyson2df90
        return dyson2df90.dyson2d(intra,tx,ty,txy,txmy,nx,ny,nk,ez)
    else:
        ns = intra.shape[0]*nx*ny
        g = np.zeros((ns,ns),dtype=np.complex_)
        nkx,nky = nk,nk
        return dyson2d_jit(intra,tx,ty,txy,txmy,nx,ny,nkx,nky,ez,g)

def dyson1d(intra,inter,nx,nkx,ez):
    """Workaround for 1D"""
    zero = intra*0.0 # zero matrix
    ns = intra.shape[0]*nx
    g = np.zeros((ns,ns),dtype=np.complex_)
    return dyson2d_jit(intra,inter,zero,zero,zero,nx,1,nkx,1,ez,g)




@jit(nopython=True)
def dyson2d_jit(intra,tx,ty,txy,txmy,nx,ny,nkx,nky,ez,g):
    """jit version of the function"""
    n = intra.shape[0] # size of the matrix
    gs = np.zeros((nkx*nky,n,n),dtype=np.complex_) # GF in k-points
    gm = np.zeros((nx*2-1,ny*2-1,n,n),dtype=np.complex_) # GF in minicells
    ks = np.zeros((nkx*nky,2)) # kpoints
    em = np.identity(n) # identity times energy
    em = em*ez # identity times energy
    ik = 0
    nk2 = nkx*nky # total number of kpoints
    # Compute all the GF
    for i in range(nkx):
        kx = 1./nkx*np.pi*2*i # kpoint
        for j in range(nky):
            ky = 1./nky*np.pi*2*j # kpoint
            m = np.exp(1j*kx)*tx + np.exp(1j*ky)*ty
            m = m + np.exp(1j*(kx+ky))*txy + np.exp(1j*(kx-ky))*txmy
            m = m + np.conjugate(m.T) + intra # Bloch Hamiltonian
            gs[ik,:,:] = np.linalg.inv(em - m) # store GF
            ks[ik,0] = kx
            ks[ik,1] = ky
            ik += 1 # increase counter
    # GF in the different "minicells" of the supercell
    for i in range(-nx+1,nx):
        for j in range(-ny+1,ny):
            m = np.zeros((n,n),dtype=np.complex_) # initialize
            for ik in range(nk2): # loop over kpoints
                m = m + gs[ik,:,:]*np.exp(1j*(ks[ik,0]*i+ks[ik,1]*j))
            gm[nx+i-1,ny+j-1,:,:] = m[:,:]/nk2 # store
    # now create indexes for the minicell
    inds = np.zeros((nx*ny,2),dtype=np.int_)
    ic = 0
    for i in range(nx):
        for j in range(ny):
            inds[ic,0] = i
            inds[ic,1] = j
            ic += 1
    # now compute the supercell GF
    for i in range(nx*ny): # loop over rows
        i1 = inds[i,0] # minicell index
        j1 = inds[i,1] # minicell index
        for j in range(nx*ny):
            i2 = inds[j,0] # minicell index
            j2 = inds[j,1] # minicell index
            ii = i1-i2 + nx -1 
            jj = j1-j2 + ny -1
            m[:,:] = gm[ii,jj,:,:] # get this matrix
            ii0 = n*i
            ii1 = n*(i+1)
            jj0 = n*j
            jj1 = n*(j+1)
            g[ii0:ii1,jj0:jj1] = m[:,:] # store all this data
    return g









import numpy as np
from ..green import green_renormalization
from ..algebra import dagger


def get_selfenergy(self,energy,lead=0,delta=None,pristine=False):
   """Return self energy of iesim lead"""
   # in case you use a dummy selfenergy
   if self.use_minimal_selfenergy:
       return minimal_selfenergy(self,lead=lead)
   if delta is None:  delta = self.delta
# if the interpolation function has been created
   if self.interpolated_selfenergy:
       raise # this is not maintained anymore
       if lead==0: return np.matrix(self.selfgen[0](energy)) # return selfenergy
       if lead==1: return np.matrix(self.selfgen[1](energy)) # return selfenergy
# run the calculation
   else:
       if lead==0:
           intra = self.left_intra
           inter = self.left_inter
           if pristine: cou = self.left_inter
           else: cou = self.left_coupling*self.scale_lc
           deltal = delta + self.extra_delta_left # new delta left
       elif lead==1:
           intra = self.right_intra
           inter = self.right_inter
           if pristine: cou = self.right_inter
           else: cou = self.right_coupling*self.scale_rc
           deltal = delta + self.extra_delta_right # new delta right
       else: raise # not implemented
       ggg,gr = green_renormalization(intra,inter,energy=energy,delta=deltal)
       selfr = cou@gr@dagger(cou) # selfenergy
       return selfr # return selfenergy





def minimal_selfenergy(self,lead=0,**kwargs):
    """Function returning a minimal selfenergy"""
    print("Using a dummy metallic selfenergy")
    if not self.use_minimal_selfenergy: raise # make sure this is ok
    gamma = self.minimal_selfenergy_gamma # get the gamma
    # reescale for compatibility with everything else
    if lead==0: 
        scale = self.scale_lc**2
        n = self.left_inter.shape[0]
    elif lead==1: 
        scale = self.scale_rc**2
        n = self.right_inter.shape[0]
    out = np.identity(n,dtype=np.complex_) # output matrix
    out = -scale*gamma*out*1j # multiply
    return out




from __future__ import print_function
import numpy as np


def anderson(h,w=0.0,p=1.0):
  """Return a Hamiltonian with Anderson disorder"""
  def fdis(r):
      if np.random.random()<p: # probability of an impurity
        return (np.random.random() - .5)*2*w # disorder
      return 0.0
  h.add_onsite(fdis)
  return h


def phase(h,w=0.0):
  """Random phase disorder"""
  if h.has_eh: raise # not yet
  if h.has_spin: raise # not yet
  ho = h.copy() # copy the Hamiltonian
  n = h.intra.shape[0] # dimension
  cs = np.array(np.random.random((n,n)) - .5)*2*w # disorder
  cs = (cs - cs.T)/2. # Hermitian
  cs = np.exp(1j*np.pi*cs) # phases
  ho.intra = np.matrix(np.array(ho.intra)*np.array(cs))
  return ho





# library for fractal geometries
import numpy as np

def sierpinski(n=4,mode="triangle"):
    """Create a sierpinski geometry"""
    from .. import supercell
    from ..geometry import triangular_lattice
    from .. import geometry
    g = triangular_lattice() # get geometry
    if mode=="triangular":
        r = np.array([[0.,0.,0.]])
        a1 = np.array([1.,0.,0.])
        a2 = np.array([np.cos(np.pi/3),np.sin(np.pi/3),0.])
        vs = [a1,a2]
        inflate = lambda i: 2**i # inflate function
    elif mode=="square":
        r = np.array([[0.,0.,0.]])
        a1 = np.array([1.,0.,0.])
        a2 = np.array([0.,1.,0.])
        vs = [a1,2*a1,a2,2*a2,2*a1+a2,a1+2*a2,2*a2+2*a1]
        inflate = lambda i: 3**i # inflate function
    elif mode=="honeycomb":
        g0 = geometry.honeycomb_lattice_C6()
#        g0 = geometry.kagome_lattice() #honeycomb_lattice_C6()
#        g0 = geometry.honeycomb_lattice()
        g0 = supercell.target_angle_volume(g0,angle=1./3.,volume=1,
                                           same_length=True)
#        g0 = g0.get_supercell([2,2])
#        g0.dimensionality = 0
#        g0 = g0.clean()
        r = g0.r
        a1 = g0.a1
        a2 = g0.a2 
        vs = [a1,a2]
        inflate = lambda i: 2**i # inflate function
    else: raise
    for i in range(n): # loop over replicas
        r0 = r.copy() # copy
        for a in vs:
            r = np.concatenate([r,r0[:]+inflate(i)*a]) # add positions
    go = g.copy()
    go.dimensionality = 0
    go.r = r
    go.r2xyz()
    return go # return positions




import numpy as np
from . import fullgreen
from ..checkclass import is_iterable

def get_dos(self,energies=None,write=True,nk=20,**kwargs):
    """Compute density of states"""
    if self.dimensionality==1:
        if energies is None: energies = np.linspace(-1.0,1.0,100)
        ds = [self.get_coupled_central_dos(energy=e,**kwargs) for e in energies]
        if write: np.savetxt("DOS.OUT",np.array([energies,ds]).T)
        return energies,np.array(ds)
    if self.dimensionality==2:
        def fun(k,e):
            if self.dimensionality==2: # 2D heterostructure
                HT1 = self.generate(k) # generate heterostructure
                return HT1.get_coupled_central_dos(energy=e,**kwargs)
        ks = np.linspace(0.,1.,nk) # kpoints
        ds = [] # empty list
        for e in energies:
            ds.append(np.mean([fun(k,e) for k in ks]))
        return (energies,ds)
    else: raise





def device_dos(HT,energy=0.0,mode="central",
        operator=None,ic=0,**kwargs):
   """ Calculates the density of states 
       of a HTstructure by a  
       green function approach, input is a HTstructure class
   """
   g = fullgreen.get_full_green(HT,energy,mode=mode,ic=ic,**kwargs)
   if ic is None or is_iterable(ic): # all the cells
       g0 = 0. # initialize
       for i in range(len(g)): # loop
         gi = g[i]
         if operator is not None:
           gi = HT.Hr.get_operator(operator)*gi
         g0 = g0 + gi
       g = g0 # overwrite
   else: # specific cell, assume a single matrix has been returned
     if operator is not None:
       g = HT.Hr.get_operator(operator)*g
   d = -np.trace(g.imag)
   return d




# library to perform embedding calculations
from __future__ import print_function

from . import green
from . import parallel
from . import algebra
from . import timing
import numpy as np
import scipy.linalg as lg
import os
from .increase_hilbert import full2profile
from . import filesystem as fs
from .hamiltonians import Hamiltonian


class Embedding():
    """Define an embedding object"""
    def __init__(self,h,m=None,nsuper=None,selfenergy=None):
        self.H = h.copy() # Pristine Hamiltonian
        self.geometry = self.H.geometry
        self.dimensionality = self.H.dimensionality
        self.selfenergy = selfenergy 
        self.has_gf_generator = False
        self.has_spin = self.H.has_spin
        self.has_eh = self.H.has_eh
        self.boundary_embedding_generator = None
        self.mode = "bulk" # mode
        self.nsuper = None # supercell between original Hamiltonian
        if m is not None: 
            if type(m)==Hamiltonian: 
#                print("Picking intracell in embedding")
                m = m.intra # get the intracell
            self.m = m # provided matrix
            if m.shape[0]!=h.intra.shape[0]: 
                if nsuper is None:
                    print("Dimensions do not match in embedding")
                    raise
                else: 
                    self.nsuper = nsuper # store the supercell
                    hs = h.supercell(nsuper)
                    if m.shape[0]!=hs.intra.shape[0]: 
                        print("Dimensions do not match after supercell")
                        print(m.shape[0],hs.intra.shape[0])
                        raise
            else: pass
        else: self.m = h.intra.copy() # pristine one
    def get_gf(self,**kwargs):
        return get_gf(self,**kwargs)
    def get_density_matrix(self,**kwargs):
        return get_dm(self,**kwargs)
    def get_ldos(self,**kwargs):
        return get_ldos(self,**kwargs)
    def ldos(self,**kwargs): return self.get_ldos(**kwargs)
    def dos(self,**kwargs):
        (x,y,d) = self.ldos(**kwargs)
        return np.sum(d) # sum the DOS
    def get_dos(self,**kwargs): return self.dos(**kwargs)
    def multidos(self,es=np.linspace(-1.0,1.0,30),
                   energies=None,**kwargs):
        if energies is not None: es = energies # overwrite
        ds = parallel.pcall(lambda e: self.dos(energy=e,**kwargs),es)
        return (es,np.array(ds))
    def multildos(self,es=np.linspace(-2.,2.,20),**kwargs):
        """Compute the ldos at different energies"""
        fs.rmdir("MULTILDOS")
        fs.mkdir("MULTILDOS")
        ds = [] # total DOS
        fo = open("MULTILDOS/MULTILDOS.TXT","w")
        # parallel execution
        out = parallel.pcall(lambda x: self.ldos(energy=x,**kwargs),es) 
        for (e,o) in zip(es,out):
            (x,y,d) = o # extract
#            (x,y,d) = self.ldos(e=e,**kwargs) # compute LDOS
            ds.append(np.mean(d)) # total DOS
            name0 = "LDOS_"+str(e)+"_.OUT" # name
            name = "MULTILDOS/"+name0
            fo.write(name0+"\n") # name of the file
            np.savetxt(name,np.array([x,y,d]).T) # save data
        np.savetxt("MULTILDOS/DOS.OUT",np.array([es,ds]).T)
    def set_multihopping(self,mh):
        """Set a multihopping as the impurity"""
        dd = mh.get_dict() # get the dictionary
        for d in dd:
            if d not in [(0,0,0)]: pass#raise # not implemented
            else:
                self.m = dd[d] # copy intra cell matrix
    def copy(self):
        from copy import deepcopy
        return deepcopy(self)
    def get_energy_ingap_state(EO,**kwargs):
        from .embeddingtk import ingap
        return ingap.energy_ingap_state(EO,**kwargs)
    # dummy methods for compatibility
    def turn_multicell(self): pass
    def get_multicell(self): return self
    def turn_dense(self): pass
    def get_dict(self): return self.H.get_dict()
    def shift_fermi(self,mu): self.H.shift_fermi(mu)
    def get_total_energy(self,**kwargs): return 0.0
    def get_mean_field_hamiltonian(self,**kwargs):
        from .selfconsistency.embedding import hubbard_mf
        return hubbard_mf(self,**kwargs) # return Hubbard mean-field
    def get_didv(self,**kwargs):
        from .embeddingtk import didv
        return didv.get_didv(self,**kwargs)
    # for compatibility with kappa functionality
    def didv(self,**kwargs): self.get_didv(**kwargs) 
    def get_kappa(self,**kwargs):
        from .embeddingtk import kappa
        return kappa.get_kappa(self,**kwargs)






def dos_impurity(h,vc=None,energies=np.linspace(-.5,.5,20),
                   mode="adaptive",delta=0.01,nk=50,silent=True,
                   use_generator=False):
  """ Calculates the green function using the embedding technique"""
  if vc is None: vc = h.intra  # assume perfect
  iden = np.identity(h.intra.shape[0],dtype=np.complex_)
  if use_generator:
    getgreen = green.green_generator(h,nk=nk) # get the function
    def get_green(energy): return getgreen(energy,delta=delta)
  else: # use the conventional method
    def get_green(energy):
      return green.bloch_selfenergy(h,energy=energy,delta=delta,nk=nk,
                                       mode=mode)
  def pfun(energy): # function to parallelize
    g,selfe = get_green(energy) # get Green and selfenergy
    emat = iden*(energy + delta*1j)  # E +i\delta 
    gv = lg.inv(emat - vc -selfe)   # Green function of a vacancy, with selfener
    d = -np.trace(g).imag  # save DOS of the pristine
    dv = -np.trace(gv).imag  # save DOS of the defected
    if not silent: print("Done",energy)
    return [d,dv]
  out = np.array(parallel.pcall(pfun,energies)) # compute
  ds,dsv = out[:,0],out[:,1] # get the different data
  np.savetxt("DOS_PRISTINE.OUT",np.array([energies,ds]).T)
  np.savetxt("DOS_DEFECTIVE.OUT",np.array([energies,dsv]).T)
  return ds,dsv # return object





def bulk_and_surface(h1,nk=100,energies=np.linspace(-1.,1.,100),**kwargs):
  """Get the surface DOS of an interface"""
  from scipy.sparse import csc_matrix,bmat
  if h1.dimensionality==2:
      kpath = [[k,0.,0.] for k in np.linspace(0.,1.,nk)]
  else: raise
  ik = 0
  h1 = h1.get_multicell() # multicell Hamiltonian
  tr = timing.Testimator("DOS") # generate object
  dos_bulk = energies*0.0
  dos_sur = energies*0.0
  for k in kpath:
    tr.remaining(ik,len(kpath)) # generate object
    ik += 1
    outs = green.surface_multienergy(h1,k=k,energies=energies,**kwargs)
    dos_bulk += np.array([-algebra.trace(g[1]).imag for g in outs])
    dos_sur += np.array([-algebra.trace(g[0]).imag for g in outs])
  dos_bulk /= len(kpath)
  dos_sur /= len(kpath)
  np.savetxt("DOS.OUT",np.array([energies,dos_bulk,dos_sur]).T)
  return energies,dos_bulk,dos_sur




def onsite_defective_central(h,m,nsuper):
    return onsite_supercell(h,nsuper,mc=m)


def onsite_supercell(h,nsuper,mc=None):
    if h.is_multicell: h = h.get_no_multicell() # redefine
    from .checkclass import is_iterable
    if not is_iterable(nsuper): # just a number 
        if h.dimensionality==1: nsuper = [nsuper,1]
        elif h.dimensionality==2: nsuper = [nsuper,nsuper]
        else: raise
    if h.dimensionality>2: raise
    #### this is a dirty workaround ####
    if h.dimensionality==1:
        h = h.copy()
        h.tx = h.inter
        h.ty = h.intra*0.0
        h.txy = h.intra*0.0
        h.txmy = h.intra*0.0
    #### end of the dirty trick ###
    inds = []
    k = 0
    n = nsuper[0]*nsuper[1] # number of cells
    intrasuper = [[None for j in range(n)] for i in range(n)]
    for i in range(nsuper[0]):
      for j in range(nsuper[1]):
        inds += [(i,j)]
        k += 1
    from scipy.sparse import bmat
    from scipy.sparse import csc_matrix as csc
    tx = csc(h.tx)
    ty = csc(h.ty)
    txy = csc(h.txy)
    txmy = csc(h.txmy)
    intra = csc(h.intra)
    if mc is None: mc = intra
    else: mc = csc(mc)
    for i in range(n):
      intrasuper[i][i] = intra # intracell
      (x1,y1) = inds[i]
      for j in range(n):
        (x2,y2) = inds[j]
        dx = x2-x1
        dy = y2-y1
        if dx==1 and  dy==0: intrasuper[i][j] = tx
        if dx==-1 and dy==0: intrasuper[i][j] = tx.H
        if dx==0 and  dy==1: intrasuper[i][j] = ty
        if dx==0 and  dy==-1: intrasuper[i][j] = ty.H
        if dx==1 and  dy==1: intrasuper[i][j] = txy
        if dx==-1 and dy==-1: intrasuper[i][j] = txy.H
        if dx==1 and  dy==-1: intrasuper[i][j] = txmy
        if dx==-1 and dy==1: intrasuper[i][j] = txmy.H
    if nsuper[0]%2==1: # central cell
        ii=int(n/2)
        intrasuper[ii][ii] = mc # central onsite
    else:
        ii=int(n/2)
        ii = ii - int(nsuper[0]/2)
        intrasuper[ii][ii] = mc # central onsite
    intrasuper = bmat(intrasuper).todense() # supercell
    return intrasuper


def get_gf(self,**kwargs):
    """Return the Green's function"""
    if self.selfenergy is None: # no selfenergy given, compute it
        return get_gf_exact(self,**kwargs)
    else: # a selfenergy is given
        from .embeddingtk.boundaryembedding import boundary_embedding_gf
        return boundary_embedding_gf(self,selfenergy=self.selfenergy,**kwargs)


def get_gf_exact(self,energy=0.0,delta=1e-2,
        nsuper=1,nk=100,operator=None,**kwargs):
    """Return the Green's function"""
    h = self.H
    e = energy
    if self.nsuper is None: # old way
        g,selfe = green.supercell_selfenergy(h,e=e,delta=delta,nk=nk,
                nsuper=nsuper,gtype=self.mode)
    else: # workaround for supercells
        raise # this does not work yet
        g,selfe = green.supercell_selfenergy(h,e=e,delta=delta,nk=nk,
                nsuper=nsuper*self.nsuper) # compute Green's function
        h = h.supercell(self.nsuper) # and redefine with a supercell
    ms = onsite_supercell(h,nsuper)
    n = self.m.shape[0]
    ms = onsite_defective_central(h,self.m,nsuper)
    ns = ms.shape[0] # dimension of the supercell
    iden = np.identity(ns,dtype=np.complex_) # identity
    emat = iden*(e + delta*1j) # energy matrix
    gv = algebra.inv(emat - ms -selfe)   # Defective Green function
    return gv


def get_onsite(self,nsuper=1,**kwargs):
    h = self.H
    ms = onsite_defective_central(h,self.m,nsuper)
    return ms



from .embeddingtk.embedded import get_dm
from .embeddingtk.embedded import embed_hamiltonian
from .embeddingtk.selfenergies import get_selfenergy_from_potential



def get_ldos(self,energy=0.0,delta=1e-2,nsuper=1,nk=100,
                    write = True,return_rd = False,
                    operator=None,**kwargs):
    """Compute the local density of states"""
    h = self.H
    # get the Green's function
    gv = self.get_gf(energy=energy,delta=delta,nsuper=nsuper,nk=nk)
    if operator is not None:
        operator = h.get_operator(operator) # overwrite
        gv = operator*gv # multiply
    ds = [-gv[i,i].imag/np.pi for i in range(gv.shape[0])] # LDOS
    ds = full2profile(h,ds,check=False) # resum if necessary
    ds = np.array(ds) # convert to array
    gs = h.geometry.supercell(nsuper)
    x,y,z,r = gs.x,gs.y,gs.z,gs.r
    if write: np.savetxt("LDOS.OUT",np.array([x,y,ds]).T)
    if return_rd:
        return r,ds
    else:
        return x,y,ds








